# Starts experiments from command line

import inspect
import json
import logging
import pkgutil
import sys
from functools import reduce
from importlib import import_module
from pathlib import Path
from typing import List, Optional, Tuple, Union

import click
import docstring_parser
import omegaconf
from experimaestro import LauncherRegistry, RunMode, experiment
from experimaestro.settings import get_workspace
from omegaconf import OmegaConf, SCMode
from termcolor import cprint

import xpmir.papers as papers
from xpmir.configuration import omegaconf_argument
from xpmir.learning.optim import TensorboardService
from xpmir.papers.helpers import PaperExperiment
from xpmir.papers.results import PaperResults
from xpmir.experiments.ir import UploadToHub


class ExperimentsCli(click.MultiCommand):
    def __init__(
        self, pkg_name: str, experiments: List[papers.Experiment], *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.pkg_name = pkg_name
        self.id2experiment = {experiment.id: experiment for experiment in experiments}
        self.experiments = experiments

    def list_commands(self, ctx: click.Context):
        return [experiment.id for experiment in self.experiments]

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        try:
            experiment = self.id2experiment[cmd_name]
        except KeyError:
            logging.error(
                "Configuration %s not found among %s",
                cmd_name,
                " ".join(self.id2experiment.keys()),
            )
            raise
        sub_package, name = experiment.cli.split(":")
        module = import_module(f"{self.pkg_name}.{sub_package}")
        return getattr(module, name)


class PapersCli(click.MultiCommand):
    def list_commands(self, ctx):
        path = str(Path(papers.__file__).parent)
        names = []
        for pkg in pkgutil.walk_packages([path]):
            names.append(pkg.name)
        return names

    def get_command(self, ctx, name):
        pkg_name = f"{__package__}.{name}"
        try:
            mod = import_module(pkg_name)

            papers = mod.PAPERS  # type: List[papers.Experiment]
            return ExperimentsCli(pkg_name, papers)
        except AttributeError:
            pass

        return


def paper_command(
    package: Optional[str] = None,
    folder: Optional[Union[str, Path]] = None,
    schema=None,
    tensorboard_service=False,
):
    """General command line decorator for an XPM-IR experiment

    This annotation adds a set of arguments for the

    HuggingFace upload: the documentation comes from the docstring

    :param tensorboard_service: If true, register a tensorboard service and transmits it
        as ``tensorboard_service`` to function.

    .. seealso:: :py:func:`omegaconf_argument` for `folder` and `package` documentation
    """

    omegaconf_schema = None
    if schema is not None:
        omegaconf_schema = OmegaConf.structured(schema())

    def _decorate(fn):
        decorators = [
            click.command(),
            click.option("--debug", is_flag=True, help="Print debug information"),
            click.option("--show", is_flag=True, help="Print configuration and exits"),
            click.option(
                "--env",
                help="Define one environment variable",
                type=(str, str),
                multiple=True,
            ),
            click.option(
                "--host",
                type=str,
                default=None,
                help="Server hostname (default to localhost,"
                " not suitable if your jobs are remote)",
            ),
            click.option(
                "--run-mode",
                type=click.Choice(RunMode),
                default=RunMode.NORMAL,
                help="Sets the run mode",
            ),
            click.option(
                "--xpm-config-dir",
                type=Path,
                default=None,
                help="Path for the experimaestro config directory "
                "(if not specified, use $HOME/.config/experimaestro)",
            ),
            click.option(
                "--port",
                type=int,
                default=None,
                help="Port for monitoring (can be defined in the settings.yaml file)",
            ),
            click.option(
                "--upload-to-hub",
                required=False,
                type=str,
                default=None,
                help="Upload the model to Hugging Face Hub with the given identifier",
            ),
            click.option(
                "--workdir",
                type=str,
                default=None,
                help="Working directory - if None, uses the default XPM "
                "working directory",
            ),
            omegaconf_argument(
                "--configuration",
                package=package,
                folder=folder,
                click_mode=click.option,
                required=True,
            ),
            click.argument("args", nargs=-1, type=click.UNPROCESSED),
        ]

        def cli(
            show,
            debug,
            configuration,
            workdir: Path,
            host: Optional[str],
            port: Optional[int],
            xpm_config_dir: Optional[Path],
            env: List[Tuple[str, str]],
            args,
            run_mode: RunMode,
            upload_to_hub: Optional[str],
            **kwargs,
        ):
            nonlocal omegaconf_schema
            assert schema is None or omegaconf_schema is not None

            logging.getLogger().setLevel(logging.DEBUG if debug else logging.INFO)
            logging.getLogger("xpm.hash").setLevel(logging.INFO)
            conf_args = OmegaConf.from_dotlist(args)

            if xpm_config_dir is not None:
                assert xpm_config_dir.is_dir()
                LauncherRegistry.set_config_dir(xpm_config_dir)

            configuration: PaperExperiment = OmegaConf.merge(configuration, conf_args)
            if omegaconf_schema is not None:
                try:
                    configuration: PaperExperiment = OmegaConf.merge(
                        omegaconf_schema, configuration
                    )
                except omegaconf.errors.ConfigKeyError as e:
                    cprint(f"Error in configuration:\n\n{e}", "red", file=sys.stderr)
                    sys.exit(1)

            if show:
                print(json.dumps(OmegaConf.to_container(configuration)))  # noqa: T201
                sys.exit(0)

            configuration = OmegaConf.to_container(
                configuration, structured_config_mode=SCMode.INSTANTIATE
            )
            parameters = inspect.signature(fn).parameters

            if upload_to_hub is not None:
                if configuration.title == "" and configuration.description == "":
                    doc = docstring_parser.parse(fn.__doc__)
                else:
                    doc = f"# {configuration.title}\n{configuration.description}"
                upload_to_hub = UploadToHub(upload_to_hub, doc)
                kwargs["upload_to_hub"] = upload_to_hub

            kwargs = {**kwargs, "debug": debug, "run_mode": run_mode}

            kwargs = {key: value for key, value in kwargs.items() if key in parameters}

            # Run the experiment
            if run_mode == RunMode.NORMAL:
                logging.info("Starting experimaestro server (%s:%s)", host, port)

            if workdir is None or not Path(workdir).is_dir():
                workdir = get_workspace(workdir).path.expanduser().resolve()
                logging.info("Using working directory %s", workdir)

            with experiment(
                workdir, configuration.id, host=host, port=port, run_mode=run_mode
            ) as xp:
                if tensorboard_service:
                    kwargs["tensorboard_service"] = xp.add_service(
                        TensorboardService(xp.resultspath / "runs")
                    )

                for key, value in env:
                    xp.setenv(key, value)

                results = fn(xp, configuration, **kwargs)
                xp.wait()

                if isinstance(results, PaperResults) and run_mode == RunMode.NORMAL:
                    if upload_to_hub is not None and "upload_to_hub" not in parameters:
                        upload_to_hub.send_scorer(
                            results.models,
                            evaluations=results.evaluations,
                            tb_logs=results.tb_logs,
                        )

                    print(results.evaluations.to_dataframe())  # noqa: T201
                return results

        cli.__doc__ = fn.__doc__
        cmd = reduce(lambda fn, decorator: decorator(fn), decorators, cli)
        return cmd

    return _decorate


papers_cli = PapersCli(help="Runs an experiment from a paper")



from functools import cached_property as attrs_cached_property  # noqa: F401
from experimaestro.experiments import configuration, ConfigurationBase  # noqa: F401

# Renaming for compatibility
Experiment = ConfigurationBase



from functools import cached_property
from collections import OrderedDict, defaultdict
from typing import ClassVar, Dict, Iterator, List, Tuple, Any
import torch
import numpy as np
from datamaestro.record import Record, record_type
from datamaestro_text.data.ir import (
    create_record,
    DocumentStore,
    InternalIDItem,
    SimpleTextItem,
    TextItem,
)

from experimaestro import Param
from xpmir.text.encoders import TextEncoder, RepresentationOutput


class SampleDocumentStore(DocumentStore):
    id: Param[str] = ""
    num_docs: Param[int] = 200

    def __post_init__(self):
        # Generate all the documents
        self.documents = OrderedDict(
            (
                str(ix),
                create_record(
                    InternalIDItem(ix),
                    id=str(ix),
                    text=f"Document {ix}",
                ),
            )
            for ix in range(self.num_docs)
        )

    @property
    def documentcount(self):
        return len(self.documents)

    def document_int(self, internal_docid: int) -> Record:
        return self.documents[str(internal_docid)]

    def document_ext(self, docid: str) -> Record:
        """Returns the text of the document given its id"""
        return self.documents[docid]

    def iter_documents(self) -> Iterator[Record]:
        return iter(self.documents.values())

    @cached_property
    def document_recordtype(self):
        return record_type(InternalIDItem, SimpleTextItem)

    def docid_internal2external(self, docid: int):
        """Converts an internal collection ID (integer) to an external ID"""
        return str(docid)

    def __iter__(self):
        return iter(self.documents.values())


class VectorGenerator:
    def __init__(self, dimension, sparsity):
        self.dimension = dimension
        self.num_zeros_rate = torch.FloatTensor([sparsity * self.dimension])

    def __call__(self) -> torch.Tensor:
        x = torch.randn(self.dimension)
        if self.num_zeros_rate.item() == 0:
            return x

        zeros = int(min(torch.poisson(self.num_zeros_rate).item(), self.dimension - 1))
        x[np.random.choice(np.arange(self.dimension), zeros)] = 0
        x = x.abs()
        return x


def check_str(x: Any):
    assert isinstance(x, str), f"{type(x)} is not a string"
    return x


class SparseRandomTextEncoder(TextEncoder):
    # A default dict to always return the same embeddings
    MAPS: ClassVar[Dict[Tuple[int, float], Dict[str, torch.Tensor]]] = {}

    map: Dict[str, torch.Tensor]
    dim: Param[int]
    sparsity: Param[float] = 0.0

    def __post_init__(self):
        super().__init__()
        if not (self.dim, self.sparsity) in SparseRandomTextEncoder.MAPS:
            SparseRandomTextEncoder.MAPS[(self.dim, self.sparsity)] = defaultdict(
                VectorGenerator(self.dim, self.sparsity)
            )

        # Get the given map
        self.map = SparseRandomTextEncoder.MAPS[(self.dim, self.sparsity)]

    @property
    def dimension(self):
        return self.dim

    def forward(self, texts: List[str]) -> torch.Tensor:
        """Returns a matrix encoding the provided texts"""

        tensors = [self.map[check_str(
            text[TextItem].text if isinstance(text, Record) else text
        )].unsqueeze(0) for text in texts]

        return RepresentationOutput(torch.cat(tensors))



import asyncio
from functools import cached_property
from pathlib import Path
import contextlib
import json
import logging
import os
import re
import subprocess
import sys
from typing import List, Optional
from experimaestro import tqdm as xpmtqdm, Task, Meta, field, PathGenerator

from datamaestro_text.data.ir import (
    DocumentStore,
    TextItem,
    IDItem,
    TopicRecord,
    create_record,
    InternalIDItem,
)
import datamaestro_text.data.ir.csv as ir_csv
from datamaestro_text.data.ir.trec import (
    Documents,
    Topics,
    TipsterCollection,
    TrecTopics,
)
from experimaestro import Param, progress
from tqdm import tqdm
from xpmir.index.anserini import Index
from xpmir.rankers import Retriever, ScoredDocument, document_cache
from xpmir.rankers.standard import BM25, QLDirichlet, Model
from xpmir.utils.utils import Handler, StreamGenerator, needs_java

pyserini_java = needs_java(11)


def anserini_classpath():
    import pyserini

    base = Path(pyserini.__file__).parent

    paths = [path for path in base.rglob("anserini-*-fatjar.jar")]
    if not paths:
        raise Exception(f"No matching jar file found in {base}")

    latest = max(paths, key=os.path.getctime)
    return latest


def javacommand():
    """Returns the start of the java command including the Anserini class path"""
    from jnius_config import get_classpath

    command = ["{}/bin/java".format(os.environ["JAVA_HOME"]), "-cp"]
    command.append(":".join(get_classpath() + [str(anserini_classpath())]))

    return command


@pyserini_java
class IndexCollection(Index, Task):
    """An [Anserini](https://github.com/castorini/anserini) index"""

    CLASSPATH = "io.anserini.index.IndexCollection"

    documents: Param[Documents]
    """The documents to index"""

    threads: Meta[int] = 8
    """Number of threads when indexing"""

    path: Meta[Path] = field(default_factory=PathGenerator("index"))

    id: Param[str] = ""
    """Use an empty ID since identifier is determined by documents"""

    def execute(self):
        command = javacommand()
        command.append(IndexCollection.CLASSPATH)
        command.extend(["-index", self.path, "-threads", self.threads])

        chandler = Handler()

        @chandler()
        def trec_collection(documents: TipsterCollection):
            return contextlib.nullcontext("void"), [
                "-collection",
                "TrecCollection",
                "-input",
                documents.path,
            ]

        @chandler()
        def csv_collection(documents: ir_csv.Documents):
            def _generator(out):
                counter = 0
                size = os.path.getsize(documents.path)
                with (
                    documents.path.open("rt", encoding="utf-8") as fp,
                    tqdm(total=size, unit="B", unit_scale=True) as pb,
                ):
                    for ix, line in enumerate(fp):
                        # Update progress
                        ll = len(line)
                        pb.update(ll)
                        counter += ll
                        progress(counter / size)

                        # Generate document
                        docid, text = line.strip().split(documents.separator, 1)
                        json.dump({"id": docid, "contents": text}, out)
                        out.write("\n")

            generator = StreamGenerator(_generator, mode="wt")

            return generator, [
                "-collection",
                "JsonCollection",
                "-input",
                generator.filepath.parent,
            ]

        @chandler.default()
        def generic_collection(documents: Documents):
            """Generic collection handler, supposes that we can iterate documents"""

            def _generator(out):
                logging.info(
                    "Starting the iterator over the documents (%d documents)",
                    documents.documentcount,
                )
                for document in xpmtqdm(
                    documents.iter(), unit="documents", total=documents.documentcount
                ):
                    # Generate document
                    json.dump(
                        {
                            "id": document[IDItem].id,
                            "contents": document[TextItem].text,
                        },
                        out,
                    )
                    out.write("\n")

            generator = StreamGenerator(_generator, mode="wt")

            return generator, [
                "-collection",
                "JsonCollection",
                "-input",
                generator.filepath.parent,
            ]

        generator, args = chandler[self.documents]
        command.extend(args)

        if self.storePositions:
            command.append("-storePositions")
        if self.storeDocvectors:
            command.append("-storeDocvectors")
        if self.storeRaw:
            command.append("-storeRawDocs")
        if self.storeContents:
            command.append("-storeContents")

        # Index and keep track of progress through regular expressions
        RE_FILES = re.compile(
            rb""".*index\.IndexCollection \(IndexCollection.java:\d+\) - ([\d,]+) files found"""  # noqa: E501
        )
        RE_FILE = re.compile(
            rb""".*index\.IndexCollection\$LocalIndexerThread \(IndexCollection.java:\d+\).* docs added."""  # noqa: E501
        )
        RE_COMPLETE = re.compile(
            rb""".*(IndexCollection|AbstractIndexer).*"""
            rb"""Indexing Complete.*documents indexed"""
        )

        async def run(command):
            with generator as _:
                logging.info("Running with command %s", command)
                proc = await asyncio.create_subprocess_exec(
                    *command, stderr=None, stdout=asyncio.subprocess.PIPE
                )

                nfiles = -1
                indexedfiles = 0
                complete = False

                while True:
                    data = await proc.stdout.readline()

                    if not data:
                        break

                    m = RE_FILES.match(data)
                    complete = complete or (RE_COMPLETE.match(data) is not None)
                    if m:
                        nfiles = int(m.group(1).decode("utf-8").replace(",", ""))
                    elif RE_FILE.match(data):
                        indexedfiles += 1
                        progress(indexedfiles / nfiles)
                    else:
                        sys.stdout.write(
                            data.decode("utf-8"),
                        )

                await proc.wait()

                if proc.returncode == 0 and not complete:
                    logging.error(
                        "Did not see the indexing complete log message"
                        " -- exiting with error"
                    )
                    sys.exit(1)
                sys.exit(proc.returncode)

        asyncio.run(run([str(s) for s in command]))


@pyserini_java
class SearchCollection(Task):
    index: Param[Index]
    topics: Param[Topics]
    model: Param[Model]
    path: Meta[Path] = field(default_factory=PathGenerator("results.trec"))

    def execute(self):
        command = javacommand()
        command.append("io.anserini.search.SearchCollection")
        command.extend(("-index", self.index.path, "-output", self.path))

        # Topics

        topicshandler = Handler()

        @topicshandler()
        def trectopics(topics: TrecTopics):
            return ("-topicreader", "Trec", "-topics", topics.path)

        @topicshandler()
        def tsvtopics(topics: ir_csv.Topics):
            return ("-topicreader", "TsvInt", "-topics", topics.path)

        command.extend(topicshandler[self.topics])

        # Model

        modelhandler = Handler()

        @modelhandler()
        def handle(bm25: BM25):
            return ("-bm25", "-bm25.k1", str(bm25.k1), "-bm25.b", str(bm25.b))

        command.extend(modelhandler[self.model])

        # Start
        logging.info("Starting command %s", command)
        p = subprocess.run(command)
        sys.exit(p.returncode)


@pyserini_java
class AnseriniRetriever(Retriever):
    """An Anserini-based retriever

    Attributes:
        index: The Anserini index
        model: the model used to search. Only supports BM25 so far.
        k: Number of results to retrieve
    """

    index: Param[Index]
    model: Param[Model]
    k: Param[int] = 1500

    @cached_property
    def searcher(self):
        from pyserini.search.lucene import LuceneSearcher

        searcher = LuceneSearcher(str(self.index.path))

        modelhandler = Handler()

        @modelhandler()
        def handle_bm25(bm25: BM25):
            searcher.set_bm25(bm25.k1, bm25.b)

        @modelhandler()
        def handle_qld(qld: QLDirichlet):
            searcher.set_qld(qld.mu)

        modelhandler[self.model]

        return searcher

    def _get_store(self) -> Optional[Index]:
        """Returns the associated index (if any)"""
        if self.index.storeContents:
            return self.index
        return self.store

    def retrieve(self, record: TopicRecord) -> List[ScoredDocument]:
        # see
        # https://github.com/castorini/anserini/blob/master/src/main/java/io/anserini/search/SimpleSearcher.java
        hits = self.searcher.search(record[TextItem].text, k=self.k)
        store = self.get_store()

        # Batch retrieve documents
        if store is not None:
            return [
                ScoredDocument(doc, hit.score)
                for hit, doc in zip(
                    hits, store.documents_ext([hit.docid for hit in hits])
                )
            ]

        return [
            ScoredDocument(
                create_record(
                    InternalIDItem(hit.lucene_docid),
                    id=hit.docid,
                    text=getattr(hit, "contents", None),
                ),
                hit.score,
            )
            for hit in hits
        ]


@document_cache
def index_builder(
    documents: Documents, *, launcher=None, **index_params
) -> IndexCollection:
    return IndexCollection(documents=documents, **index_params).submit(
        launcher=launcher
    )


def retriever(
    index_builder: IndexCollection,
    documents: Documents,
    *,
    k: int = None,
    model: Model = None,
    store: DocumentStore = None,
):
    """Function to construct an Anserini retriever"""
    index = index_builder(documents)

    return AnseriniRetriever(
        index=index, k=k or AnseriniRetriever.k, model=model, store=store
    )



from typing import Iterable, Iterator, List, Optional, Set, Union, Type
from pathlib import Path
from experimaestro import (
    Param,
    Config,
    Task,
    tqdm,
    pathgenerator,
    Annotated,
    Meta,
)
from experimaestro.compat import cached_property
from datamaestro_text.data.ir import (
    IDItem,
    TextItem,
    Adhoc,
    AdhocAssessments,
    DocumentRecord,
    DocumentStore,
    Documents,
    Topics,
    TopicRecord,
)

from datamaestro_text.data.ir.trec import TrecAdhocAssessments
from datamaestro_text.data.ir.csv import Topics as CSVTopics
from xpmir.rankers import Retriever
from xpmir.misc import IDList, FileIDList
from xpmir.utils.utils import easylog

logger = easylog()


class AbstractTopicFold(Topics):
    """ID-based topic selection"""

    topics: Param[Topics]
    """The collection of the topics"""

    def iter(self):
        ids = set(self.ids)
        for topic in self.topics.iter():
            if topic[IDItem].id in ids:
                yield topic

    @property
    def topic_recordtype(self) -> Type[TopicRecord]:
        """The class for topics"""
        return self.topics.topic_recordtype

    def __validate__(self) -> None:
        super().__validate__()
        assert self.topics.topic_recordtype.has(IDItem)


class TopicFold(AbstractTopicFold):
    ids: Param[List[str]]
    """A set of the ids for the topics where we select from"""


class IDTopicFold(AbstractTopicFold):
    id_list: Param[IDList]

    @property
    def ids(self):
        return self.id_list.ids


class AbstractAdhocAssessmentFold(AdhocAssessments):
    """Filter assessments by topic ID"""

    qrels: Param[AdhocAssessments]
    """The collection of the assessments"""

    def iter(self):
        ids = set(self.ids)
        for qrels in self.qrels.iter():
            if qrels.topic_id in ids:
                yield qrels


class IDAdhocAssessmentFold(AbstractAdhocAssessmentFold):
    id_list: Param[IDList]

    @property
    def ids(self):
        return self.id_list.ids


class AdhocAssessmentFold(AbstractAdhocAssessmentFold):
    """Filter assessments by topic ID"""

    ids: Param[List[str]]
    """A set of the ids for the assessments where we select from"""


def fold(ids: Iterable[str], dataset: Adhoc):
    """Returns a fold of a dataset, given topic ids"""
    ids = sorted(list(ids))
    topics = TopicFold(topics=dataset.topics, ids=ids)
    qrels = AdhocAssessmentFold(assessments=dataset.assessments, ids=ids)
    return Adhoc(topics=topics, assessments=qrels, documents=dataset.documents)


class ConcatFold(Task):
    """
    Concatenation of several datasets to get a full dataset.
    """

    datasets: Param[List[Adhoc]]
    """The list of Adhoc datasets to concatenate"""

    assessments: Annotated[Path, pathgenerator("assessments.tsv")]
    """Generated assessments file"""

    topics: Annotated[Path, pathgenerator("topics.tsv")]
    """Generated topics file"""

    def task_outputs(self, dep) -> Adhoc:
        dataset_document_id = set(dataset.document.id for dataset in self.datasets)
        assert (
            len(dataset_document_id) == 1
        ), "At the moment only one set of documents supported."
        return Adhoc(
            id="",  # No need to have a more specific id since it is generated
            topics=dep(CSVTopics(id="", path=self.topics)),
            assessments=dep(TrecAdhocAssessments(id="", path=self.assessments)),
            documents=self.datasets[0].documents,
        )

    def execute(self):
        topics = []
        # concat the topics
        for dataset in self.datasets:
            topics.extend([topic for topic in dataset.topics.iter()])

        # Write topics and assessments
        ids = set()
        self.topics.parent.mkdir(parents=True, exist_ok=True)
        with self.topics.open("wt") as fp:
            for topic in topics:
                ids.add(topic[IDItem].id)
                slash_t = "\t"
                fp.write(
                    f"""{topic[IDItem].id}\t"""
                    f"""{topic[TextItem].text.replace(slash_t, ' ')}\n"""
                )

        with self.assessments.open("wt") as fp:
            for dataset in self.datasets:
                for qrels in dataset.assessments.iter():
                    if qrels.topic_id in ids:
                        for qrel in qrels.assessments:
                            fp.write(
                                f"""{qrels.topic_id} 0 {qrel.doc_id} {qrel.rel}\n"""
                            )


class RandomFold(Task):
    """Extracts a random subset of topics from a dataset"""

    seed: Param[int]
    """Random seed used to compute the fold"""

    sizes: Param[List[float]]
    """Number of topics of each fold (or percentage if sums to 1)"""

    dataset: Param[Adhoc]
    """The Adhoc dataset from which a fold is extracted"""

    fold: Param[int]
    """Which fold should be taken"""

    exclude: Param[Optional[Topics]]
    """Exclude some topics from the random fold"""

    assessments: Annotated[Path, pathgenerator("assessments.tsv")]
    """Generated assessments file"""

    topics: Annotated[Path, pathgenerator("topics.tsv")]
    """Generated topics file"""

    def __validate__(self):
        assert self.fold < len(self.sizes)

    @staticmethod
    def folds(
        seed: int,
        sizes: List[float],
        dataset: Param[Adhoc],
        exclude: Param[Optional[Topics]] = None,
        submit=True,
    ):
        """Creates folds

        Parameters:

        - submit: if true (default), submits the fold tasks to experimaestro
        """

        folds = []
        for ix in range(len(sizes)):
            fold = RandomFold.C(
                seed=seed, sizes=sizes, dataset=dataset, exclude=exclude, fold=ix
            )
            if submit:
                fold = fold.submit()
            folds.append(fold)

        return folds

    def task_outputs(self, dep) -> Adhoc:
        return dep(
            Adhoc.C(
                id="",  # No need to have a more specific id since it is generated
                topics=dep(CSVTopics.C(id="", path=self.topics)),
                assessments=dep(TrecAdhocAssessments.C(id="", path=self.assessments)),
                documents=self.dataset.documents,
            )
        )

    def execute(self):
        import numpy as np

        # Get topics
        badids = (
            set(topic[IDItem].id for topic in self.exclude.iter())
            if self.exclude
            else set()
        )
        topics = [
            topic
            for topic in self.dataset.topics.iter()
            if topic[IDItem].id not in badids
        ]
        random = np.random.RandomState(self.seed)
        random.shuffle(topics)

        # Get the fold
        sizes = np.array([0.0] + self.sizes)
        s = sizes.sum()
        if abs(s - 1) < 1e-6:
            sizes = np.round(len(topics) * sizes)
            sizes = np.round(len(topics) * sizes / sizes.sum())

        assert sizes[self.fold + 1] > 0

        indices = sizes.cumsum().astype(int)
        topics = topics[indices[self.fold] : indices[self.fold + 1]]

        # Write topics and assessments
        ids = set()
        self.topics.parent.mkdir(parents=True, exist_ok=True)
        with self.topics.open("wt") as fp:
            for topic in topics:
                ids.add(topic[IDItem].id)
                fp.write(f"""{topic[IDItem].id}\t{topic[TextItem].text}\n""")

        with self.assessments.open("wt") as fp:
            for qrels in self.dataset.assessments.iter():
                if qrels.topic_id in ids:
                    for qrel in qrels.assessments:
                        fp.write(f"""{qrels.topic_id} 0 {qrel.doc_id} {qrel.rel}\n""")


class TopicsFoldGenerator(FileIDList, Task):
    """Extracts a random subset of topics from a dataset

    This task is more generic than the RandomFold one and should work
    whatever the topics/assessments as long as they are serializable (using
    pickle).
    """

    seed: Param[int]
    """Random seed used to compute the fold"""

    sizes: Param[List[float]]
    """Number of topics of each fold (or percentage if sums to 1)"""

    dataset: Param[Adhoc]
    """The Adhoc dataset from which a fold is extracted"""

    fold: Param[int]
    """Which fold should be taken"""

    exclude: Param[Optional[Topics]]
    """Exclude some topics from the random fold"""

    path: Annotated[Path, pathgenerator("topics.lst")]
    """Selected topic IDs"""

    def __validate__(self):
        assert self.fold < len(self.sizes)

    @staticmethod
    def folds(
        seed: int,
        sizes: List[float],
        dataset: Param[Adhoc],
        exclude: Param[Optional[Topics]] = None,
        submit=True,
    ):
        """Creates folds

        Parameters:

        - submit: if true (default), submits the fold tasks to experimaestro
        """

        folds = []
        for ix in range(len(sizes)):
            fold = TopicsFoldGenerator.C(
                seed=seed, sizes=sizes, dataset=dataset, exclude=exclude, fold=ix
            )
            if submit:
                fold = fold.submit()
            folds.append(fold)

        return folds

    def task_outputs(self, dep) -> Adhoc:
        return dep(
            Adhoc(
                id="",  # No need to have a more specific id since it is generated
                topics=dep(
                    IDTopicFold.C(id="", topics=self.dataset.topics, id_list=self)
                ),
                assessments=dep(
                    IDAdhocAssessmentFold.C(
                        id="", qrels=self.dataset.assessments, id_list=self
                    )
                ),
                documents=self.dataset.documents,
            )
        )

    def execute(self):
        import numpy as np

        # Get topics
        badids = (
            set(topic[IDItem].id for topic in self.exclude.iter())
            if self.exclude
            else set()
        )
        topics = [
            topic[IDItem].id
            for topic in self.dataset.topics.iter()
            if topic[IDItem].id not in badids
        ]
        random = np.random.RandomState(self.seed)
        random.shuffle(topics)

        # Get the fold
        sizes = np.array([0.0] + self.sizes)
        s = sizes.sum()
        if abs(s - 1) < 1e-6:
            sizes = np.round(len(topics) * sizes)
            sizes = np.round(len(topics) * sizes / sizes.sum())

        assert sizes[self.fold + 1] > 0

        indices = sizes.cumsum().astype(int)
        topics = topics[indices[self.fold] : indices[self.fold + 1]]

        # Write topic IDs
        with self.path.open("wt") as fp:
            for topic in topics:
                fp.write(f"{topic}\n")


class DocumentSubset(Documents):
    """ID-based topic selection"""

    base: Param[DocumentStore]
    """The full document store"""

    docids_path: Meta[Path]
    """Path to the file containing the document IDs"""

    in_memory: Meta[bool] = False
    """Whether to load the dataset in memory"""

    def __len__(self):
        return len(self.docids)

    def __post_init__(self):
        super().__post_init__()
        self.cache = {}

    @property
    def documentcount(self):
        return len(self.docids)

    def document_ext(self, docid: str):
        if self.in_memory:
            if doc := self.cache.get(docid, None):
                return doc

            doc = self.base.document_ext(docid)
            self.cache[docid] = doc
            return doc

        return self.base.document_ext(docid)

    def __getitem__(self, slice: Union[int, slice]):
        docids = self.docids[slice]
        if isinstance(docids, List):
            return DocumentSubsetSlice.C(self, self.docids[slice])
        return self.document_ext(docids)

    @cached_property
    def docids(self) -> List[str]:
        # Read document IDs
        with self.docids_path.open("rt") as fp:
            return [line.strip() for line in fp]

    def iter_ids(self):
        yield from self.docids

    def iter(self) -> Iterator[DocumentRecord]:
        for docid in self.iter_ids():
            return self.base.document_ext(docid)


class DocumentSubsetSlice:
    """A slice of a `DocumentSubset`"""

    def __init__(self, subset: DocumentSubset, doc_ids: List[int]):
        self.subset = subset
        self.doc_ids = doc_ids

    def __iter__(self):
        for docid in self.doc_ids:
            yield self.subset.document_ext(docid)

    def __len__(self):
        return len(self.doc_ids)

    def __getitem__(self, ix):
        return self.subset.document_ext(self.doc_ids[ix])


class RetrieverBasedCollection(Task):
    """Buils a subset of documents based on the output of a set of retrievers
    and on relevance assessment.
    First get all the document based on the assessment then add the retrieved ones.
    """

    relevance_threshold: Param[float] = 0
    """Relevance threshold"""

    dataset: Param[Adhoc]
    """A dataset"""

    retrievers: Param[List[Retriever]]
    """Rankers"""

    keepRelevant: Param[bool] = True
    """Keep documents judged relevant"""

    keepNotRelevant: Param[bool] = False
    """Keep documents judged not relevant"""

    docids_path: Annotated[Path, pathgenerator("docids.txt")]
    """The file containing the document identifiers of the collection"""

    def __validate__(self):
        assert len(self.retrievers) > 0, "At least one retriever should be given"

    def task_outputs(self, dep) -> Adhoc:
        return Adhoc(
            id="",  # No need to have a more specific id since it is generated
            topics=self.dataset.topics,
            assessments=self.dataset.assessments,
            documents=dep(
                DocumentSubset.C(
                    id="", base=self.dataset.documents, docids_path=self.docids_path
                )
            ),
        )

    def execute(self):
        for retriever in self.retrievers:
            retriever.initialize()

        # Selected document IDs
        docids: Set[str] = set()

        topics = {t.topic_id: t for t in self.dataset.assessments.iter()}

        # Retrieve all documents
        for topic in tqdm(
            self.dataset.topics.iter(), total=self.dataset.topics.count()
        ):
            qrels = topics.get(topic[IDItem].id)
            if qrels is None:
                logger.warning(
                    "Skipping topic %s [%s], (no assessment)",
                    topic[IDItem].id,
                    topic[TextItem].text,
                )
                continue

            # Add (not) relevant documents
            if self.keepRelevant:
                docids.update(
                    a.doc_id
                    for a in qrels.assessments
                    if a.rel > self.relevance_threshold
                )

            if self.keepNotRelevant:
                docids.update(
                    a.doc_id
                    for a in qrels.assessments
                    if a.rel <= self.relevance_threshold
                )

            # Retrieve and add
            # already defined the numbers to retrieve inside the retriever, so
            # don't need to worry about the threshold here
            for retriever in self.retrievers:
                docids.update(
                    sd.document[IDItem].id for sd in retriever.retrieve(topic)
                )

        # Write the document IDs
        with self.docids_path.open("wt") as fp:
            fp.writelines(f"{docid}\n" for docid in docids)


class TextStore(Config):
    """Associates an ID with a text"""

    def __getitem__(self, key: str) -> str:
        raise NotImplementedError()


class MemoryTopicStore(TextStore):
    """View a set of topics as a (in memory) text store"""

    topics: Param[Topics]
    """The collection of the topics to build the store"""

    @cached_property
    def store(self):
        return {topic[IDItem].id: topic[TextItem].text for topic in self.topics.iter()}

    def __getitem__(self, key: str) -> str:
        return self.store[key]



import builtins
import mock
import sys
import logging
from xpmir import version

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# --- Set build mode
# Used to notify python modules that we are building
# a documentation

builtins.__sphinx_build__ = True


# -- Project information -----------------------------------------------------

project = "Experimaestro IR (XPM-IR)"
copyright = "2023, Benjamin Piwowarski"
author = "Benjamin Piwowarski"

# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # Experimaestro extension
    "experimaestro.sphinx",
    "datamaestro.sphinx",
    # Math in docs with MathJax
    "sphinx.ext.mathjax",
    # Read The Docs theme
    "sphinx_rtd_theme",
    # Use Markdown parser
    "myst_parser",
    # auto documention
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    # Link to other documentations
    "sphinx.ext.intersphinx",
    # Google style docstrings
    "sphinx.ext.napoleon",
    # Named tuples
    "sphinx_toolbox.more_autodoc.autonamedtuple",
    # Code
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


intersphinx_mapping = {
    "datamaestro_text": (
        "https://datamaestro-text.readthedocs.io/en/latest",
        None,
    )  # /objects.inv",
}

# Autodoc options

autodoc_default_options = {
    "show-inheritance": True,
}

autodoc_mock_imports = [
    # "torch",
    "faiss",
    "pandas",
    "bs4",
    "pytorch_transformers",
    "pytrec_eval",
    "apex",
    "pytorch_lightning",
    "ir_measures",
]


def side_effect(*args, **kwargs):
    logging.error("Side effect %s / %s", args, kwargs)


class DocMock(mock.Mock):
    def __repr__(self):
        names = []
        mock = self
        while mock is not None:
            names.append(mock._mock_name)
            mock = mock._mock_parent

        return ".".join(names[::-1])


for name in [
    "torch",
    "torch.nn",
    "torch.distributed",
    "torch.optim",
    "torch.nn.functional",
    "torch.functional",
    "torch.multiprocessing",
    "torch.utils",
    "torch.utils.tensorboard",
    "torch.utils.tensorboard.writer",
    "torch.optim.lr_scheduler",
    "transformers",
]:
    sys.modules[name] = DocMock(side_effect=side_effect, name=name)

import torch.nn as nn  # noqa: E402


class TorchModule:
    to = None
    train = None


nn.Module = TorchModule



# Utility functions for MS-Marco experiments

from typing import Union
from functools import lru_cache

from experimaestro import Launcher

from datamaestro import prepare_dataset
from datamaestro_text.transforms.ir import (
    ShuffledTrainingTripletsLines,
    StoreTrainingTripletTopicAdapter,
    StoreTrainingTripletDocumentAdapter,
)
from datamaestro_text.data.ir import Documents, Adhoc

from xpmir.utils.functools import cache
from xpmir.datasets.adapters import RandomFold
from xpmir.evaluation import Evaluations, EvaluationsCollection
from xpmir.letor.samplers import TripletBasedSampler
from xpmir.datasets.adapters import MemoryTopicStore
from xpmir.letor.distillation.samplers import (
    DistillationPairwiseSampler,
    PairwiseHydrator,
)
from xpmir.letor.samplers.hydrators import SampleHydrator, PairwiseTransformAdapter

from xpmir.measures import AP, RR, P, nDCG, Success
from xpmir.papers import configuration


@configuration
class ValidationSample:
    seed: int = 123
    size: int = 500


@lru_cache
def prepare_collection(prepare_str: str) -> Union[Documents, Adhoc]:
    """Prepare a dataset and caches the result"""
    return prepare_dataset(prepare_str)


MEASURES = [AP, P @ 20, nDCG, nDCG @ 10, nDCG @ 20, RR, RR @ 10, Success @ 5]

# --- MsMarco v1


@cache
def msmarco_v1_docpairs_sampler(
    *,
    sample_rate: float = 1.0,
    sample_max: int = 0,
    launcher: "Launcher" = None,
) -> TripletBasedSampler:
    """Train sampler (deprecated: use msmarco_v1_docpairs_efficient_sampler)

    This uses shuffled pre-computed triplets from MS Marco

    :param sample_rate: Sample rate for the triplets (default 1)
    """
    topics = prepare_dataset("irds.msmarco-passage.train.queries")
    train_triples = prepare_dataset("irds.msmarco-passage.train.docpairs")
    triplets = ShuffledTrainingTripletsLines(
        seed=123,
        data=StoreTrainingTripletTopicAdapter(data=train_triples, store=topics),
        sample_rate=sample_rate,
        sample_max=sample_max,
        doc_ids=True,
        topic_ids=False,
    ).submit(launcher=launcher)

    # Adds the text to the documents
    triplets = StoreTrainingTripletDocumentAdapter(
        data=triplets, store=prepare_collection("irds.msmarco-passage.documents")
    )
    return TripletBasedSampler(source=triplets)


@cache
def msmarco_v1_docpairs_efficient_sampler(
    *,
    sample_rate: float = 1.0,
    sample_max: int = 0,
    launcher: "Launcher" = None,
    seed: int = 123,
) -> TripletBasedSampler:
    """Train sampler

    This uses shuffled pre-computed triplets from MS Marco

    :param sample_rate: Sample rate for the triplets (default 1)
    """
    topics = prepare_dataset("irds.msmarco-passage.train.queries")
    train_triples = prepare_dataset("irds.msmarco-passage.train.docpairs")
    triplets = ShuffledTrainingTripletsLines(
        seed=seed,
        data=StoreTrainingTripletTopicAdapter(data=train_triples, store=topics),
        sample_rate=sample_rate,
        sample_max=sample_max,
        doc_ids=True,
        topic_ids=False,
    ).submit(launcher=launcher)

    # Builds the sampler by hydrating documents
    sampler = TripletBasedSampler(source=triplets)
    hydrator = SampleHydrator(
        documentstore=prepare_collection("irds.msmarco-passage.documents")
    )

    return PairwiseTransformAdapter(sampler=sampler, adapter=hydrator)


@cache
def msmarco_v1_validation_dataset(cfg: ValidationSample, launcher=None):
    """Sample dev topics to get a validation subset"""
    return RandomFold(
        dataset=prepare_collection("irds.msmarco-passage.dev"),
        seed=cfg.seed,
        fold=0,
        sizes=[cfg.size],
        exclude=prepare_collection("irds.msmarco-passage.dev.small").topics,
    ).submit(launcher=launcher)


@cache
def msmarco_v1_tests(dev_test_size: int = 0):
    """MS-Marco default test collections: DL TREC 2019 & 2020 + devsmall

    devsmall can be restricted to a smaller dataset for debugging using dev_test_size
    """
    v1_devsmall_ds = prepare_collection("irds.msmarco-passage.dev.small")
    if dev_test_size > 0:
        (v1_devsmall_ds,) = RandomFold.folds(
            seed=0, sizes=[dev_test_size], dataset=v1_devsmall_ds
        )

    return EvaluationsCollection(
        msmarco_dev=Evaluations(v1_devsmall_ds, MEASURES),
        trec2019=Evaluations(
            prepare_dataset("irds.msmarco-passage.trec-dl-2019"), MEASURES
        ),
        trec2020=Evaluations(
            prepare_dataset("irds.msmarco-passage.trec-dl-2020"), MEASURES
        ),
    )


@cache
def msmarco_hofstaetter_ensemble_hard_negatives() -> DistillationPairwiseSampler:
    """Hard negatives from Hofstätter et al. (2020)

    Hard negatives trained by distillation with cross-encoder Improving
    Efficient Neural Ranking Models with Cross-Architecture Knowledge
    Distillation, (Sebastian Hofstätter, Sophia Althammer, Michael Schröder,
    Mete Sertkan, Allan Hanbury), 2020
    """
    train_triples_distil = prepare_dataset(
        "com.github.sebastian-hofstaetter." "neural-ranking-kd.msmarco.ensemble.teacher"
    )

    # Access to topic text
    train_topics = prepare_dataset("irds.msmarco-passage.train.queries")

    # Combine the training triplets with the document and queries texts
    distillation_samples = PairwiseHydrator(
        samples=train_triples_distil,
        documentstore=prepare_collection("irds.msmarco-passage.documents"),
        querystore=MemoryTopicStore(topics=train_topics),
    )

    # Generate a sampler from the samples
    return DistillationPairwiseSampler(samples=distillation_samples)


@cache
def finetuning_validation_dataset(
    cfg: ValidationSample, dataset_id: str, launcher=None
):
    """Sample dev topics to get a validation subset"""
    return RandomFold(
        dataset=prepare_collection(dataset_id),
        seed=cfg.seed,
        fold=0,
        sizes=[cfg.size],
    ).submit(launcher=launcher)



from functools import cached_property
import itertools
from experimaestro import Param
from typing import Iterator, Tuple
from datamaestro.record import record_type
import datamaestro_text.data.ir as ir
from xpmir.letor.samplers import (
    TrainingTriplets,
    TripletBasedSampler,
)
from xpmir.datasets.adapters import TextStore

from xpmir.letor.samplers.hydrators import (
    SampleHydrator,
    PairwiseTransformAdapter,
)


class TripletIterator(TrainingTriplets):
    def iter(
        self,
    ) -> Iterator[Tuple[ir.TopicRecord, ir.DocumentRecord, ir.DocumentRecord]]:
        count = 0

        while True:
            yield ir.create_record(id=str(count)), ir.create_record(
                id=str(2 * count)
            ), ir.create_record(id=str(2 * count + 1))
            count += 1

    @cached_property
    def topic_recordtype(self):
        return record_type(ir.IDItem)

    @cached_property
    def document_recordtype(self):
        return record_type(ir.IDItem)


class FakeTextStore(TextStore):
    def __getitem__(self, key: str) -> str:
        return f"T{key}"


class FakeDocumentStore(ir.DocumentStore):
    id: Param[str] = ""

    def document_ext(self, docid: str) -> ir.DocumentRecord:
        return ir.create_record(id=docid, text=f"D{docid}")


def test_pairwise_hydrator():
    sampler = TripletBasedSampler(source=TripletIterator(id="test-triplets"))

    hydrator = SampleHydrator(
        querystore=FakeTextStore(), documentstore=FakeDocumentStore()
    )

    h_sampler = PairwiseTransformAdapter(sampler=sampler, adapter=hydrator)
    h_sampler.instance()

    for record, n in zip(h_sampler.pairwise_iter(), range(5)):
        assert record.query[ir.TextItem].text == f"T{n}"
        assert record.positive[ir.TextItem].text == f"D{2*n}"
        assert record.negative[ir.TextItem].text == f"D{2*n+1}"

    batch_it = h_sampler.pairwise_batch_iter(3)
    for record, n in zip(itertools.chain(next(batch_it), next(batch_it)), range(5)):
        assert record.query[ir.TextItem].text == f"T{n}"
        assert record.positive[ir.TextItem].text == f"D{2*n}"
        assert record.negative[ir.TextItem].text == f"D{2*n+1}"



from experimaestro.experiments import configuration, ConfigurationBase
from experimaestro.launcherfinder import find_launcher

from functools import cached_property as attrs_cached_property
from xpmir.learning.devices import CudaDevice, BestDevice, Device
from xpmir.letor import Random


PaperExperiment = ConfigurationBase


@configuration()
class NeuralIRExperiment(ConfigurationBase):
    """Settings most neural IR experiments"""

    gpu: bool = True
    """Use GPU for computation"""

    use_best_device: bool = False
    """Use best GPU device"""

    seed: int = 0
    """The seed used for experiments"""

    @attrs_cached_property
    def random(self):
        return Random.C(seed=self.seed)

    @attrs_cached_property
    def device(self) -> Device:
        if self.use_best_device:
            return BestDevice.C()
        return CudaDevice.C() if self.gpu else Device.C()


@configuration()
class LauncherSpecification:
    """Launcher specification

    This allows requesting computational resources such as 2 GPUs with more than
    12Go of memory)
    """

    requirements: str

    @attrs_cached_property
    def launcher(self):
        return find_launcher(self.requirements)



import numpy as np
import atexit
from abc import ABC, abstractmethod
from queue import Full, Empty
import torch.multiprocessing as mp
from typing import (
    Generic,
    Callable,
    Dict,
    Tuple,
    List,
    Iterable,
    Iterator,
    Protocol,
    TypeVar,
    Any,
    TypedDict,
)
from xpmir.utils.utils import easylog
import logging

logger = easylog()

# --- Utility classes

State = TypeVar("State")
T = TypeVar("T")
U = TypeVar("U")


class iterable_of(Generic[T]):
    def __init__(self, factory: Callable[[], Iterator[T]]):
        self.factory = factory

    def __iter__(self):
        return self.factory()


class SerializableIterator(Iterator[T], Generic[T, State]):
    """An iterator that can be serialized through state dictionaries.

    This is used when saving the sampler state
    """

    @abstractmethod
    def state_dict(self) -> State:
        ...

    @abstractmethod
    def load_state_dict(self, state: State):
        ...


class SerializableIteratorAdapter(SerializableIterator[T, State], Generic[T, U, State]):
    """Adapts a serializable iterator with a transformation function based on
    the iterator"""

    def __init__(
        self,
        main: SerializableIterator[T, State],
        generator: Callable[[SerializableIterator[T, State]], Iterator[U]],
    ):
        self.generator = generator
        self.main = main
        self.iter = generator(main)

    def load_state_dict(self, state):
        self.main.load_state_dict(state)
        self.iter = self.generator(self.main)

    def state_dict(self):
        return self.main.state_dict()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iter)


class BatchIteratorAdapter(SerializableIterator[List[T], State]):
    """Adapts a serializable iterator into a batchwise serializable iterator"""

    def __init__(self, iterator: SerializableIterator[T, State], size: int):
        self.iterator = iterator
        self.size = size

    def state_dict(self):
        return self.iterator.state_dict()

    def load_state_dict(self, state):
        self.iterator.load_state_dict(state)

    def __iter__(self):
        return self

    def __next__(self) -> List[T]:
        batch = []
        for _, record in zip(range(self.size), self.iterator):
            batch.append(record)
        return batch


class SerializableIteratorTransform(
    SerializableIterator[T, State], Generic[T, U, State]
):
    """Adapts a serializable iterator with a transformation function"""

    def __init__(
        self,
        iterator: SerializableIterator[T, State],
        transform: Callable[[T], U],
    ):
        self.transform = transform
        self.iterator = iterator

    def load_state_dict(self, state):
        self.iterator.load_state_dict(state)

    def state_dict(self):
        return self.iterator.state_dict()

    def __iter__(self):
        return self

    def __next__(self):
        return self.transform(next(self.iterator))


class GenericSerializableIterator(SerializableIterator[T, State]):
    def __init__(self, iterator: Iterator[T]):
        self.iterator = iterator
        self.state = None

    @abstractmethod
    def state_dict(self) -> State:
        """Generate the current state dictionary"""
        ...

    @abstractmethod
    def restore_state(self, state: State):
        """Restore the iterator"""
        ...

    def load_state_dict(self, state: State):
        self.state = state

    def __next__(self):
        # Nature of the documents
        if self.state is not None:
            self.restore_state(self.state)
            self.state = None

        # And now go ahead
        return self.next()


class RandomSerializableIterator(SerializableIterator[T, Any]):
    """A serializable iterator based on a random seed"""

    def __init__(
        self,
        random: np.random.RandomState,
        generator: Callable[[np.random.RandomState], Iterator[T]],
    ):
        """Creates a new iterator based on a random generator

        Args:
            random (np.random.RandomState): The initial random state

            generator (Callable[[np.random.RandomState], Iterator[T]]): Generate
            a new iterator from a random seed
        """
        self.random = random
        self.generator = generator
        self.iter = generator(random)

    def load_state_dict(self, state):
        self.random.set_state(state["random"])
        self.iter = self.generator(self.random)

    def state_dict(self):
        return {"random": self.random.get_state()}

    def __next__(self):
        return next(self.iter)


class RandomizedSerializableIteratorState(TypedDict):
    random: Dict[str, Any]
    state: Any


class RandomStateSerializableIterator(SerializableIterator[T, State], ABC):
    @abstractmethod
    def set_random(self, random: np.random.RandomState):
        ...


class RandomStateSerializableAdaptor(RandomStateSerializableIterator[T, State], ABC):
    """Adapter for random state-biased iterator"""

    def __init__(self, iterator: SerializableIterator[T, State]):
        self.random = None
        self.iterator = iterator

    def set_random(self, random: np.random.RandomState):
        self.random = random

    def load_state_dict(self, state: State):
        return self.iterator.load_state_dict(state)

    def state_dict(self) -> State:
        return self.iterator.state_dict()


class RandomizedSerializableIterator(
    RandomSerializableIterator[T, RandomizedSerializableIteratorState[State]],
    Generic[T, State],
):
    """Serializable iterator with a random state"""

    def __init__(
        self, random: np.random.RandomState, iterator: RandomStateSerializableIterator
    ):
        """Creates a new iterator based on a random generator

        Args:
            random (np.random.RandomState): The initial random state

            generator (Callable[[np.random.RandomState], Iterator[T]]): Generate
            a new iterator from a random seed
        """
        self.random = random
        self.iterator = iterator
        iterator.set_random(self.random)

    def load_state_dict(self, state: RandomizedSerializableIteratorState[State]):
        self.random.set_state(state["random"])
        self.iterator.set_random(self.random)
        self.iterator.load_state_dict(state["state"])

    def state_dict(self) -> RandomizedSerializableIteratorState[State]:
        return {"random": self.random.get_state(), "state": self.iterator.state_dict()}

    def __next__(self):
        return next(self.iterator)


class SkippingIteratorState(TypedDict):
    """Skipping iterator state"""

    count: int


class SkippingIterator(GenericSerializableIterator[T, SkippingIteratorState]):
    """An iterator that skips the first entries and can output its state

    When serialized (i.e. checkpointing), the iterator saves the current
    position. This can be used when deserialized, to get back to the same
    (checkpointed) position.
    """

    position: int
    """The current position (in number of items) of the iterator"""

    def __init__(self, iterator: Iterator[T]):
        super().__init__(iterator)
        self.position = 0

    def state_dict(self) -> SkippingIteratorState:
        return {"count": self.position}

    def restore_state(self, state: SkippingIteratorState):
        count = state["count"]
        logger.info("Skipping %d records to match state (sampler)", count)

        assert count >= self.position, "Cannot iterate backwards"
        for _ in range(count - self.position):
            next(self.iterator)
        self.position = count

    def next(self) -> T:
        self.position += 1
        return next(self.iterator)

    @staticmethod
    def make_serializable(iterator):
        if not isinstance(iterator, SerializableIterator):
            logging.info("Wrapping iterator into a skipping iterator")
            return SkippingIterator(iterator)

        return iterator


class InfiniteSkippingIterator(SkippingIterator[T, SkippingIteratorState]):
    """Subclass of the SkippingIterator that loops an infinite number of times"""

    def __init__(self, iterable: Iterable[T]):
        super().__init__(iter(iterable))
        self.iterable = iterable

    def next(self) -> T:
        try:
            return super().next()
        except StopIteration:
            self.iterator = iter(self.iterable)
            self.position = 1
            return next(self.iterator)


class StopIterationClass:
    pass


STOP_ITERATION = StopIterationClass()


def mp_iterate(iterator, queue: mp.Queue, event: mp.Event):
    try:
        while not event.is_set():
            value = next(iterator)

            while True:
                try:
                    queue.put(value, timeout=1)
                    break
                except Full:
                    if event.is_set():
                        logger.warning("Stopping as requested by the main process")
                        queue.close()
                        break

    except StopIteration:
        logger.info("Signaling that the iterator has finished")
        queue.put(STOP_ITERATION)
    except Exception as e:
        logger.exception("Exception while iterating")
        queue.put(e)

    logger.info("End of multi-process iterator")
    queue.close()


class QueueBasedMultiprocessIterator(Iterator[T]):
    """This Queue-based iterator can be pickled when a new process is spawn"""

    def __init__(self, queue: "mp.Queue[T]", stop_process: mp.Event):
        self.queue = queue
        self.stop_process = stop_process
        self.stop_iteration = mp.Event()

    def __next__(self):
        # Get the next element
        while True:
            try:
                element = self.queue.get(timeout=1)
                break
            except Empty:
                if self.stop_iteration.is_set():
                    self.stop_process.set()
                    raise StopIteration()

        # Last element
        if isinstance(element, StopIterationClass):
            # Just in case
            self.stop_process.set()
            self.stop_iteration.set()
            raise StopIteration()

        # An exception occurred
        elif isinstance(element, Exception):
            self.stop_iteration.set()
            self.stop_process.set()
            raise RuntimeError("Error in iterator process") from element

        return element


class MultiprocessIterator(Iterator[T]):
    def __init__(self, iterator: Iterator[T], maxsize=100):
        self.process = None
        self.maxsize = maxsize
        self.iterator = iterator
        self.stop_process = mp.Event()
        self.mp_iterator = None

    def start(self):
        """Start the iterator process"""
        if self.process is None:
            self.queue = mp.Queue(self.maxsize)
            self.process = mp.Process(
                target=mp_iterate,
                args=(self.iterator, self.queue, self.stop_process),
                daemon=True,
            )

            # Start the process
            self.process.start()
            self.mp_iterator = QueueBasedMultiprocessIterator(
                self.queue, self.stop_process
            )

            atexit.register(self.close)
        return self

    def close(self):
        if self.mp_iterator:
            atexit.unregister(self.close)
            self.stop_process.set()
            self.mp_iterator = None
            logging.info("Signaled the mp_iterator to quit")

    def detach(self):
        """Produces an iterator only based on the multiprocess queue (useful
        when using torch mp.spawn)"""
        self.start()
        return self.mp_iterator

    def __next__(self):
        # Start a process if needed
        self.start()
        try:
            return next(self.mp_iterator)
        except StopIteration:
            atexit.unregister(self.close)


class StatefullIterator(Iterator[Tuple[T, State]], Protocol[State]):
    """An iterator that iterate over tuples (value, state)"""

    def load_state_dict(self, state: State):
        ...


class StatefullIteratorAdapter(Iterator[T], Generic[T, State]):
    """Adapts a serializable iterator a stateful iterator that iterates over
    (value, state) pairs"""

    def __init__(self, iterator: SerializableIterator[T, State]):
        self.iterator = iterator

    def __next__(self):
        value = next(self.iterator)
        state = self.iterator.state_dict()
        return value, state


class MultiprocessSerializableIterator(
    MultiprocessIterator[T], SerializableIterator[T, State]
):
    """A multi-process adapter for serializable iterators

    This can be used to obtain a multiprocess iterator from a serializable iterator
    """

    def __init__(self, iterator: SerializableIterator[T, State], maxsize=100):
        super().__init__(StatefullIteratorAdapter(iterator), maxsize=maxsize)

    def state_dict(self) -> Dict:
        return self.state

    def load_state_dict(self, state):
        assert self.process is None, "The iterator has already been used"
        self.iterator.iterator.load_state_dict(state)
        self.state = state

    def __next__(self):
        value, self.state = super().__next__()
        return value



import logging
from datamaestro_text.data.ir.base import (
    IDItem,
    TopicRecord,
    create_record,
)
from experimaestro import Param
from datamaestro_text.data.ir.trec import AdhocRun
from xpmir.rankers import Retriever, ScoredDocument


class AdhocRunRetriever(Retriever):
    """Retrieves using a run file"""

    run: Param[AdhocRun]
    """The run"""

    @property
    def runs(self) -> dict[str, list[ScoredDocument]]:
        result = {}
        for query_id, scored_documents in self.run.get_dict().items():
            scored_documents = list(
                [
                    ScoredDocument(create_record(id=doc_id), score)
                    for doc_id, score in scored_documents.items()
                ]
            )
            scored_documents.sort(key=lambda x: x.score, reverse=True)
            result[query_id] = scored_documents

        return result

    def retrieve(self, record: TopicRecord) -> list[ScoredDocument]:
        scored_documents = self.runs.get(record[IDItem].id, [])
        if len(scored_documents) == 0:
            logging.warning("No documents for topic %s", record[IDItem].id)
        return scored_documents



from pathlib import Path
from typing import List, Tuple
from experimaestro.compat import cached_property
from experimaestro import Choices, Param, Annotated
from datamaestro_text.data.ir import AdhocIndex


class Index(AdhocIndex):
    """Anserini-backed index

    Attributes:

        path: Path to the index
        storePositions: Store term positions
        storeDocvectors: Store document term vectors
        storeRaw: Store raw document
        storeContents: Store processed documents (e.g. without HTML tags)
        stemmer: The stemmer to use
    """

    path: Param[Path]
    storePositions: Param[bool] = False
    storeDocvectors: Param[bool] = False
    storeRaw: Param[bool] = False
    storeContents: Param[bool] = False
    stemmer: Annotated[str, Choices(["porter", "krovetz", "none"])] = "porter"

    _index_reader = None
    _stats = None

    @cached_property
    def index_reader(self):
        from pyserini.index import IndexReader

        return IndexReader(str(self.path))

    def __getstate__(self):
        return {
            key: value for key, value in self.__dict__.items() if key != "index_reader"
        }

    @cached_property
    def documentcount(self) -> int:
        return self.index_reader.stats()["documents"]

    @cached_property
    def termcount(self):
        return self.index_reader.stats()["total_terms"]

    def document_text(self, docid):
        doc = self.index_reader.doc(docid)
        return doc.contents() if doc is not None else None

    @cached_property
    def terms(self):
        """Returns a map"""
        return {entry.term: (entry.df, entry.cf) for entry in self.index_reader.terms()}

    def iter_documents(self) -> List[Tuple[str, str]]:
        """Returns the document contents"""
        for i in range(self.documentcount):
            docid = self.index_reader.convert_internal_docid_to_collection_docid(i)
            yield docid, self.document_text(docid)

    def docid_internal2external(self, ix):
        return self.index_reader.convert_internal_docid_to_collection_docid(ix)

    def term_df(self, term: str):
        x: List[str] = self.index_reader.analyze(term)
        if x:
            return self.terms.get(x[0], (0, 0))[0]
        return 0



from typing import List, Optional

import numpy as np

from datamaestro_text.data.ir import DocumentStore

from xpmir.letor.records import DocumentRecord
from xpmir.letor.samplers import Sampler
from experimaestro import Param

from xpmir.utils.iter import RandomSerializableIterator
from xpmir.utils.utils import easylog
import logging

logger = easylog()

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class MLMSampler(Sampler):
    """Sample texts from various sources

    This sampler can be used for Masked Language Modeling to sample from several
    datasets.
    """

    datasets: Param[List[DocumentStore]]
    """Lists of datasets to sample from"""

    _stores: List[DocumentStore]

    def initialize(self, random: Optional[np.random.RandomState]):
        super().initialize(random)

    def record_iter(self) -> RandomSerializableIterator[DocumentRecord]:
        def iter(random: np.random.RandomState):
            while True:
                # We could imagine setting weights here to give more importance
                # to one dataset
                document_count = [dataset.documentcount for dataset in self.datasets]
                choice = random.randint(0, len(self.datasets))
                if document_count[choice] < 10_000_000:
                    document = self.datasets[choice].document_int(
                        self.random.randint(0, self.datasets[choice].documentcount)
                    )
                    yield document
                else:
                    # FIXME: it makes the iter not fully serializable
                    yield from self.datasets[choice].iter()

        return RandomSerializableIterator(self.random, iter)



from abc import ABC, abstractmethod
from typing import Optional, TypeVar, Sequence, Iterator, Iterable
import numpy as np
from functools import cached_property
from experimaestro import Config, Param
from xpmir.utils.utils import EasyLogger
from xpmir.utils.iter import SerializableIterator, BatchIteratorAdapter


class Random(Config):
    """Random configuration"""

    seed: Param[int] = 0
    """The seed to use so the random process is deterministic"""

    @cached_property
    def state(self) -> np.random.RandomState:
        return np.random.RandomState(self.seed)

    def __getstate__(self):
        return {"seed": self.seed}


class Sampler(Config, EasyLogger):
    """Abstract data sampler"""

    def initialize(self, random: Optional[np.random.RandomState]):
        self.random = random or np.random.RandomState(random.randint(0, 2**31))


T = TypeVar("T")


class SampleIterator(Config, Iterable[T], ABC):
    """Generic class to iterate over items or batch of items"""

    @abstractmethod
    def __iter__() -> Iterator[T]:
        pass

    def __batch_iter__(self, batch_size: int) -> Iterator[Sequence[T]]:
        """Batch iterations"""
        iterator = self.__iter__()
        data = []
        try:
            while True:
                data.append(next(iterator))
                if len(data) == batch_size:
                    yield data
        except StopIteration:
            pass

        if data:
            yield data


class BaseSampler(Sampler, SampleIterator[T], ABC):
    """A serializable sampler iterator"""

    @abstractmethod
    def __iter__() -> SerializableIterator[T]:
        pass

    def __batch_iter__(self, batch_size: int) -> BatchIteratorAdapter:
        return BatchIteratorAdapter(self.__iter__(), batch_size)



import logging
import os
from abc import ABC, abstractmethod
from dataclasses import InitVar
from pathlib import Path
from typing import Any, Tuple, Type

import torch.nn as nn
from experimaestro import Config, Param

from xpmir.learning import Module
from xpmir.learning.optim import ModuleInitMode, ModuleInitOptions
from xpmir.text import TokenizedTexts

try:
    from transformers import AutoConfig, AutoModel, AutoModelForMaskedLM
except Exception:
    logging.error("Install huggingface transformers to use these configurations")
    raise


class HFModelConfig(Config, ABC):
    """Base class for all HuggingFace model configurations"""

    @abstractmethod
    def __call__(
        self,
        options: ModuleInitOptions,
        autoconfig: Type[AutoModel],
        automodel: Type[AutoConfig],
    ) -> Tuple[Any, Any]:
        """Returns a configuration and a model

        :param options: The initiatization options
        :param autoconfig: configuration factory
        :param automodel: model factory
        :return: a Tuple (configuration, model)
        """
        ...


class HFModelConfigFromId(HFModelConfig):
    model_id: Param[str]
    """HuggingFace Model ID"""

    def get_config(
        self,
        options: ModuleInitOptions,
        autoconfig: Type[AutoModel],
        automodel: Type[AutoConfig],
    ):
        model_id_or_path = self.model_id

        # Use saved models
        if model_path := os.environ.get("XPMIR_TRANSFORMERS_CACHE", None):
            path = (
                Path(model_path)
                / Path(f"{automodel.__module__}.{automodel.__qualname__}")
                / Path(self.model_id)
            )
            if path.is_dir():
                logging.warning("Using saved model from %s", path)
                model_id_or_path = path
            else:
                logging.warning(
                    "Could not find saved model in %s, using HF loading", path
                )

        # Load the model configuration
        config = autoconfig.from_pretrained(model_id_or_path, trust_remote_code=True)

        # Return it
        return config, model_id_or_path

    def __call__(
        self,
        options: ModuleInitOptions,
        autoconfig: Type[AutoConfig],
        automodel: Type[AutoModel],
    ):
        config, model_id_or_path = self.get_config(options, autoconfig, automodel)

        if options.mode == ModuleInitMode.NONE or options.mode == ModuleInitMode.RANDOM:
            logging.info("Random initialization of HF model")
            return config, automodel.from_config(config, trust_remote_code=True)

        logging.info(
            "Loading model from HF (%s) with model %s.%s",
            self.model_id,
            automodel.__module__,
            automodel.__name__,
        )
        return config, automodel.from_pretrained(model_id_or_path, config=config, trust_remote_code=True)


class HFModel(Module):
    """Base transformer class from Huggingface

    The config specifies the architecture
    """

    config: Param[HFModelConfig]
    """Model ID from huggingface"""

    model: InitVar[AutoModel]
    """The HF model"""

    @classmethod
    def from_pretrained_id(cls, model_id: str):
        return cls.C(config=HFModelConfigFromId.C(model_id=model_id))

    @property
    def autoconfig(self):
        return AutoConfig

    @property
    def automodel(self):
        return AutoModel

    def __initialize__(self, options: ModuleInitOptions):
        """Initialize the HuggingFace transformer

        Args:
            options: loader options
        """
        super().__initialize__(options)

        self.hf_config, self.model = self.config(
            options, self.autoconfig, self.automodel
        )

    @property
    def contextual_model(self) -> nn.Module:
        """Returns the model that only outputs base representations"""

        # This method needs to be updated to cater for various types of models,
        # i.e. MLM, classification, etc.
        return self.model

    def forward(self, tokenized: TokenizedTexts):
        tokenized = tokenized.to(self.model.device)
        kwargs = {}
        if tokenized.token_type_ids is not None:
            kwargs["token_type_ids"] = tokenized.token_type_ids

        return self.model(
            input_ids=tokenized.ids,
            attention_mask=tokenized.mask,
        )


class HFMaskedLanguageModel(HFModel):
    model: InitVar[AutoModelForMaskedLM]

    @property
    def automodel(self):
        return AutoModelForMaskedLM



from experimaestro import experiment
from datamaestro import prepare_dataset
from xpmir.measures import AP, nDCG
from xpmir.interfaces.anserini import IndexCollection, AnseriniRetriever
from xpmir.rankers.standard import BM25
from xpmir.evaluation import Evaluate
import os
import logging

logging.basicConfig(level=logging.INFO)
collection = prepare_dataset("irds.antique.train")

with experiment("workdir", "evaluate-bm25", port=12345) as xp:
    # Build the index
    xp.setenv("JAVA_HOME", os.environ["JAVA_HOME"])
    index = IndexCollection(documents=collection.documents).submit()

    bm25_retriever = AnseriniRetriever(k=1500, index=index, model=BM25())
    bm25_eval = Evaluate(
        dataset=collection, retriever=bm25_retriever, measures=[AP, nDCG @ 10]
    ).submit()

print("BM25 results")
print(bm25_eval.results.read_text())



from xpmir.neural.generative.hf import T5ConditionalGenerator, LoadFromT5
from xpmir.neural.generative import BeamSearchGenerationOptions
from xpmir.test import skip_if_ci
from transformers import T5ForConditionalGeneration, T5Tokenizer
import torch


@skip_if_ci
@torch.no_grad()
def test_t5_generate():
    """Test consistency in generation"""

    hf_id = "t5-small"
    # our model
    options = BeamSearchGenerationOptions(num_return_sequences=5, num_beams=5)
    test_text = ["test one", "test two"]
    our_model: T5ConditionalGenerator = T5ConditionalGenerator(
        hf_id=hf_id,
    ).instance()
    LoadFromT5(t5_model=our_model).execute()
    our_model.eval()
    our_output = our_model.generate(test_text, options)
    our_sequence = our_output.sequences

    # official model
    tokenizer = T5Tokenizer.from_pretrained(hf_id)
    official_model = T5ForConditionalGeneration.from_pretrained(hf_id)
    official_model.eval()

    input = tokenizer(
        test_text,
        max_length=tokenizer.model_max_length,
        truncation=True,
        padding=True,
        return_tensors="pt",
        return_attention_mask=True,
    )

    official_output = official_model.generate(
        input_ids=input["input_ids"],
        attention_mask=input["attention_mask"],
        max_new_tokens=10,
        num_beams=5,
        num_return_sequences=5,
        return_dict_in_generate=True,
        output_scores=True,
    )
    official_sequences = official_output.sequences

    assert torch.equal(our_sequence, official_sequences)



from typing import Dict, Iterator
from experimaestro import Param
import numpy as np
from xpmir.learning.context import (
    TrainerContext,
)
from . import Trainer


class MultipleTrainer(Trainer):
    """This trainer can be used to combine various trainers"""

    trainers: Param[Dict[str, Trainer]]
    """The trainers"""

    def initialize(
        self,
        random: np.random.RandomState,
        context: TrainerContext,
    ):
        super().initialize(random, context)
        for trainer in self.trainers.values():
            trainer.initialize(random, context)

    def load_state_dict(self, state: Dict):
        for key, trainer in self.trainers.items():
            trainer.load_state_dict(state[key])

    def state_dict(self):
        return {key: trainer.state_dict() for key, trainer in self.trainers.items()}

    def to(self, device):
        """Change the computing device (if this is needed)"""
        super().to(device)
        for trainer in self.trainers.values():
            trainer.to(device)

    def iter_batches(self) -> Iterator:
        iters = {key: trainer.iter_batches() for key, trainer in self.trainers.items()}
        while True:
            yield {key: next(iter) for key, iter in iters.items()}

    def process_batch(self, batch):
        for key, trainer in self.trainers.items():
            with self.context.scope(key):
                trainer.process_batch(batch[key])



from typing import Optional, Union
from pathlib import Path
from importlib import resources
from omegaconf import OmegaConf
from experimaestro.click import click


class OmegaConfParamType(click.Choice):
    def __init__(self, package, folder, names):
        super().__init__(names)
        self.package = package
        self.folder = folder

    def convert(self, value: str, param, ctx):
        # Load the YAML file
        if self.folder:
            return OmegaConf.load(self.folder / f"{value}.yaml")
        with resources.path(self.package, f"{value}.yaml") as path:
            return OmegaConf.load(path)


def omegaconf_argument(
    name: str,
    *,
    package: Optional[str] = None,
    folder: Optional[Union[str, Path]] = None,
    click_mode=click.argument,
    **kwargs,
):
    """Provides a choice of YAML configuration (file names)

    :param name: the name for the parameter
    :param folder: The folder containing YAML files (it can be either a Path or
        a string, and either a folder or a file within the folder)

    :param package: the module qualified name in which to search for yaml files
    """
    names = []  # list of available configurations

    YAML_SUFFIX = ".yaml"

    assert (
        folder is None or package is None
    ), "folder and package cannot be used at the same time"

    if folder is not None:
        if isinstance(folder, str):
            folder = Path(folder)
        if folder.is_file():
            folder = folder.parent

        for config_path in folder.glob(f"*{YAML_SUFFIX}"):
            names.append(config_path.name[: -len(YAML_SUFFIX)])

    if package is not None:
        for item in resources.contents(package):
            if resources.is_resource(package, item) and item.endswith(YAML_SUFFIX):
                names.append(item[: -len(YAML_SUFFIX)])

    def handler(method):
        return click_mode(
            name, type=OmegaConfParamType(package, folder, names), **kwargs
        )(method)

    return handler



from abc import ABC, abstractmethod
from typing import List, NamedTuple, Optional, TypeVar, Generic
from attr import define
import re

import torch

from experimaestro import Config
from xpmir.text.utils import lengthToMask
from xpmir.learning.optim import ModuleInitOptions
from xpmir.utils.utils import Initializable
from xpmir.utils.misc import opt_slice
from xpmir.utils.torch import to_device


class TokenizedTexts(NamedTuple):
    """Tokenized texts output"""

    tokens: Optional[List[List[str]]]
    """The list of tokens"""

    ids: torch.LongTensor
    """A matrix containing the token IDs"""

    lens: List[int]
    """the lengths of each text (in tokens)"""

    mask: Optional[torch.LongTensor]
    """The mask for the ids matrix"""

    token_type_ids: Optional[torch.LongTensor] = None
    """Type of each token"""

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, ix):
        return TokenizedTexts(
            [opt_slice(self.tokens, i) for i in ix],
            self.ids[ix],
            [self.lens[i] for i in ix],
            opt_slice(self.mask, ix),
            opt_slice(self.token_type_ids, ix),
        )

    def to(self, device: torch.device):
        if device is self.ids.device:
            return self

        return TokenizedTexts(
            self.tokens,
            self.ids.to(device),
            self.lens,
            to_device(self.mask, device),
            to_device(self.token_type_ids, device),
        )

    def subset(self, ids: torch.Tensor) -> "TokenizedTexts":
        """Returns a subset of the tokenized texts
        Arguments:
            ids: a boolean mask or a list of indices to select
        """
        length = torch.sum(ids).item() if isinstance(ids, torch.Tensor) else len(ids)
        return TokenizedTexts(
            [self.tokens[i] for i in ids] if self.tokens else None,
            self.ids[:,ids],
            [length for i in self.lens],
            self.mask[:,ids] if self.mask is not None else None,
            self.token_type_ids[:,ids] if self.token_type_ids is not None else None,
        )

class Tokenizer(Config):
    """
    Represents a vocabulary and a tokenization method

    **Deprecated**: Use TokenizerBase instead
    """

    def tokenize(self, text: str):
        """
        Meant to be overwritten in to provide vocab-specific tokenization when necessary
        e.g., BERT's WordPiece tokenization
        """
        text = text.lower()
        text = re.sub(r"[^a-z0-9]", " ", text)
        return text.split()

    def pad_sequences(self, tokensList: List[List[int]], batch_first=True, maxlen=0):
        padding_value = 0
        lens = [len(s) for s in tokensList]
        if maxlen is None:
            maxlen = max(lens)
        else:
            maxlen = min(maxlen or 0, max(lens))

        if batch_first:
            out_tensor = torch.full(
                (len(tokensList), maxlen), padding_value, dtype=torch.long
            )
            for i, tokens in enumerate(tokensList):
                out_tensor[i, : lens[i], ...] = torch.LongTensor(tokens[:maxlen])
        else:
            out_tensor = torch.full(
                (maxlen, len(tokensList)), padding_value, dtype=torch.long
            )
            for i, tokens in enumerate(tokensList):
                out_tensor[: lens[i], i, ...] = tokens[:maxlen]

        return out_tensor.to(self._dummy_params.device), lens

    def batch_tokenize(
        self, texts: List[str], batch_first=True, maxlen=None, mask=False
    ) -> TokenizedTexts:
        """
        Returns tokenized texts

        Arguments:
            mask: Whether a mask should be computed
        """
        toks = [self.tokenize(text) for text in texts]
        tokids, lens = self.pad_sequences(
            [[self.tok2id(t) for t in tok] for tok in toks],
            batch_first=batch_first,
            maxlen=maxlen,
        )

        _mask = lengthToMask(torch.LongTensor(lens)) if mask else None

        return TokenizedTexts(toks, tokids, lens, _mask)

    @property
    def pad_tokenid(self) -> int:
        raise NotImplementedError()

    def tok2id(self, tok: str) -> int:
        """
        Converts a token to an integer id
        """
        raise NotImplementedError()

    def id2tok(self, idx: int) -> str:
        """
        Converts an integer id to a token
        """
        raise NotImplementedError()

    def lexicon_size(self) -> int:
        """
        Returns the number of items in the lexicon
        """
        raise NotImplementedError()


TokenizerInput = TypeVar("TokenizerInput")
TokenizerOutput = TypeVar("TokenizerOutput", bound=TokenizedTexts)


@define
class TokenizerOptions:
    max_length: Optional[int] = None
    return_mask: Optional[bool] = True
    return_length: Optional[bool] = True


class TokenizerBase(
    Config, Initializable, Generic[TokenizerInput, TokenizerOutput], ABC
):
    """Base tokenizer"""

    def __initialize__(self, options: ModuleInitOptions):
        super().__initialize__(options)

    @abstractmethod
    def tokenize(
        self, inputs: TokenizerInput, options: Optional[TokenizerOptions] = None
    ) -> TokenizerOutput:
        """Encodes the inputs"""
        ...

    @abstractmethod
    def vocabulary_size(self) -> int:
        """
        Returns the number of tokens
        """
        ...

    @abstractmethod
    def tok2id(self, tok: str) -> int:
        """
        Converts a token to an integer id
        """
        ...

    @abstractmethod
    def id2tok(self, idx: int) -> str:
        """
        Converts an integer id to a token
        """
        ...



def test_pointwise():
    # TODO: implement tests for trainers
    pass


def test_pairwise():
    # TODO: implement tests for trainers
    pass


def test_batchwise():
    # TODO: implement tests for trainers
    pass



"""Index for sparse models"""

from abc import abstractmethod, ABC
import asyncio
from functools import cached_property
import logging
import shutil
import threading
import heapq
import torch
from queue import Empty, Queue
import torch.multiprocessing as mp
import numpy as np
import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Generic, Iterator, Union, Any
from attrs import define
from experimaestro import (
    Config,
    Task,
    Param,
    Meta,
    field,
    PathGenerator,
    tqdm,
    Constant,
)
from datamaestro_text.data.ir import DocumentRecord, DocumentStore
from xpmir.learning import ModuleInitMode
from xpmir.learning.batchers import Batcher
from xpmir.utils.utils import batchiter, easylog
from xpmir.letor import Device, DeviceInformation, DEFAULT_DEVICE
from xpmir.text.encoders import TextEncoderBase, TextsRepresentationOutput, InputType
from xpmir.rankers import Retriever, TopicRecord, ScoredDocument
from xpmir.utils.iter import MultiprocessIterator
from xpmir.utils.multiprocessing import StoppableQueue, available_cpus
import impact_index

logger = easylog()

# --- Index and retriever


@define(frozen=True)
class EncodedDocument:
    docid: int
    value: torch.Tensor


@define(frozen=True)
class DocumentRange:
    rank: int
    start: int
    end: int

    def __lt__(self, other: "DocumentRange"):
        return self.start < other.start


class DocumentIterator:
    def __init__(
        self, documents: DocumentStore, last_doc_id: None | int, max_docs, batch_size
    ):
        self.documents = documents
        self.last_doc_id = last_doc_id
        self.max_docs = max_docs
        self.batch_size = batch_size

    @cached_property
    def iterator(self):
        start = 0 if self.last_doc_id is None else self.last_doc_id + 1
        iter = self.documents.iter_documents_from(start)

        return batchiter(
            self.batch_size,
            zip(
                range(start, self.max_docs or sys.maxsize),
                iter,
            ),
        )

    def __next__(self):
        return next(self.iterator)


class AbstractSparseRetrieverIndexBuilder(Task, ABC, Generic[InputType]):
    """Builds an index from a sparse representation

    Assumes that document and queries have the same dimension, and
    that the score is computed through an inner product
    """

    documents: Param[DocumentStore]
    """Set of documents to index"""

    encoder: Param[TextEncoderBase[InputType, TextsRepresentationOutput]]
    """The encoder"""

    batcher: Meta[Batcher] = field(default_factory=Batcher.C)
    """Batcher used when computing representations"""

    batch_size: Param[int]
    """Size of batches"""

    ordered_index: Param[bool]
    """Ordered index: if not ordered, use DAAT strategy (WAND), otherwise, use
    fast top-k strategies"""

    device: Meta[Device] = DEFAULT_DEVICE
    """The device for building the index"""

    version: Constant[int] = 3
    """Version 3 of the index"""

    max_docs: Param[int] = 0
    """Maximum number of indexed documents"""

    def execute(self):
        if mp.get_start_method(allow_none=True) is None:
            mp.set_start_method("spawn")

        max_docs = 0
        if self.max_docs:
            max_docs = min(self.max_docs, self.documents.documentcount or sys.maxsize)
            logger.warning("Limited indexing to %d documents", max_docs)

        self.encoder.initialize(ModuleInitMode.DEFAULT.to_options(None))

        closed = mp.Event()
        queues = [
            StoppableQueue(2 * self.batch_size + 1, closed)
            for _ in range(self.device.n_processes)
        ]

        # Start the index process (thread)
        last_doc_id_queue = Queue()
        index_thread = threading.Thread(
            target=self.index,
            name="index",
            args=(queues, max_docs, last_doc_id_queue),
        )
        index_thread.start()

        last_doc_id = last_doc_id_queue.get()
        logger.info(f"Starting to index {max_docs} documents")
        if last_doc_id:
            logging.info(f" -- recovering from last indexed document: {last_doc_id}")

        iter_batches = MultiprocessIterator(
            DocumentIterator(self.documents, last_doc_id, max_docs, self.batch_size)
        ).detach()

        try:
            self.device.execute(
                self.device_execute,
                iter_batches,
                self.encoder,
                self.batcher,
                self.batch_size,
                queues,
            )
        except Exception:
            logger.exception("Got an exception while running encoders")

        finally:
            logger.info("Waiting for the index process to stop")
            index_thread.join()
            if not self.index_done:
                raise RuntimeError("Indexing thread did not complete")

    def index(
        self,
        queues: List[StoppableQueue[Union[DocumentRange, EncodedDocument]]],
        max_docs: int,
        last_doc_id_queue: Queue[Optional[int]],
    ):
        """Index encoded documents

        :param queues: Queues are used to send tensors
        """
        self.index_done = False
        with tqdm(
            total=max_docs,
            unit="documents",
            desc="Building the index",
        ) as pb:
            try:
                # Get ranges
                logger.info(
                    "Starting the indexing process (%d queues) in %s",
                    len(queues),
                    self.index_path,
                )

                if last_doc_id := self.create_index_builder():
                    logger.info("Starting back from document %d", last_doc_id)
                    pb.update(last_doc_id)

                logging.info("Last indexed document: %s", last_doc_id)
                last_doc_id_queue.put(last_doc_id)

                heap = [queue.get() for queue in queues]
                heapq.heapify(heap)

                # Loop over them
                while heap:
                    # Process current range
                    current = heap[0]
                    logger.debug("Handling range: %s", current)
                    for docid in range(current.start, current.end + 1):
                        encoded = queues[current.rank].get()
                        assert (
                            encoded.docid == docid
                        ), f"Mismatch in document IDs ({encoded.docid} vs {docid})"

                        (nonzero_ix,) = encoded.value.nonzero()
                        self.add_encoded_document(docid, encoded, nonzero_ix)
                        pb.update()

                    # Get next range
                    # type: DocumentRange
                    next_range = queues[current.rank].get()
                    if next_range:
                        logger.debug("Got next range: %s", next_range)
                        heapq.heappushpop(heap, next_range)
                    else:
                        logger.info("Iterator %d is over", current.rank)
                        heapq.heappop(heap)

                logger.info("Building the index")
                self.build_index()

                logger.info("Index built")
                self.index_done = True
            except Empty:
                logger.warning("One encoder got a problem... stopping")
                raise
            except BaseException:
                # Close all the queues
                logger.exception(
                    "Got an exception in the indexing process, closing the queues"
                )
                queues[0].stop()
                raise

    @staticmethod
    def device_execute(
        device_information: DeviceInformation,
        iter_batches: Iterator[List[Tuple[int, DocumentRecord]]],
        encoder,
        batcher,
        batch_size,
        queues: List[StoppableQueue],
    ):
        try:
            # Encode all documents
            logger.info(
                "Load the encoder and "
                f"transfer to the target device {device_information.device}"
            )

            encoder = encoder.to(device_information.device).eval()
            queue = queues[device_information.rank]
            batcher = batcher.initialize(batch_size)

            # Index
            with torch.no_grad():
                for batch in iter_batches:
                    # Signals the output range
                    document_range = DocumentRange(
                        device_information.rank, batch[0][0], batch[-1][0]
                    )
                    logger.debug(
                        "Starting range [%d] %s",
                        device_information.rank,
                        document_range,
                    )
                    queue.put(document_range)

                    # Outputs the documents
                    batcher.process(
                        batch,
                        SparseRetrieverIndexBuilder.encode_documents,
                        encoder,
                        queue,
                    )

            # Build the index
            logger.info("Closing queue %d", device_information.rank)
            queue.put(None)
        except BaseException:
            logging.exception("Got an exception while encoding documents")
            queue.stop()
            raise

    @staticmethod
    def encode_documents(
        batch: List[Tuple[int, DocumentRecord]],
        encoder: TextEncoderBase[InputType, TextsRepresentationOutput],
        queue: "Queue[EncodedDocument]",
    ):
        # Assumes for now dense vectors
        vectors = encoder([d for _, d in batch]).value.cpu().numpy()  # bs * vocab
        for vector, (docid, _) in zip(vectors, batch):
            queue.put(EncodedDocument(docid, vector))

    @abstractmethod
    def build_index(self):
        ...

    @abstractmethod
    def add_encoded_document(self, docid, encoded, nonzero_ix):
        ...

    @abstractmethod
    def create_index_builder(self) -> int | None:
        """Creates a new index builder

        :return: the last doc ID when using checkpointing (so we start back with
            the next document)
        """
        ...


class SparseRetriever(Retriever, Generic[InputType]):
    index: Param["AbstractSparseRetrieverIndex"]
    """The sparse retriever index"""

    encoder: Param[TextEncoderBase[InputType, TextsRepresentationOutput]]
    """Encodes InputType records to text representation output"""

    topk: Param[int]
    """Number of documents to return"""

    device: Meta[Device] = DEFAULT_DEVICE
    """The device for building the index"""

    batcher: Meta[Batcher] = field(default_factory=Batcher.C)
    """The way to prepare batches of queries (when using retrieve_all)"""

    batchsize: Meta[int]
    """Size of batches (when using retrieve_all)"""

    in_memory: Meta[bool] = False
    """Whether the index should be fully loaded in memory (otherwise, uses
    virtual memory)"""

    def initialize(self):
        super().initialize()
        logger.info("Initializing the encoder")
        self.encoder.initialize(ModuleInitMode.DEFAULT.to_options(None))
        self.encoder.to(self.device.value)
        logger.info("Initializing the index")
        self.index.initialize(self.in_memory)

    def retrieve_all(
        self, queries: Dict[str, InputType]
    ) -> Dict[str, List[ScoredDocument]]:
        """Input queries: {id: text}"""

        async def aio_search_worker(progress, results: Dict, queue: asyncio.Queue):
            try:
                while True:
                    key, query, topk = await queue.get()
                    results[key] = await self.index.aio_retrieve(
                        query, topk, **self.get_retrieval_options()
                    )
                    progress.update(1)
                    queue.task_done()
            except asyncio.exceptions.CancelledError:
                # Just stopped
                pass
            except Exception:
                logger.exception("Error in worker thread")

        async def reducer(
            batch: List[Tuple[str, InputType]],
            queue: asyncio.Queue,
        ):
            x = self.encoder([topic for _, topic in batch]).value.cpu().detach().numpy()
            assert len(x) == len(batch), (
                f"Discrepancy between counts of vectors ({len(x)})"
                f" and number queries ({len(batch)})"
            )
            for (key, _), vector in zip(batch, x):
                (ix,) = vector.nonzero()
                query = {ix: float(v) for ix, v in zip(ix, vector[ix])}
                logger.debug("Adding topic %s to the queue", key)
                await queue.put((key, query, self.topk))
                logger.debug("[done] Adding topic %s to the queue", key)

        async def aio_process():
            workers = []
            results = {}
            try:
                queue = asyncio.Queue(available_cpus())
                items = list(queries.items())

                with tqdm(
                    desc="Retrieve documents", total=len(items), unit="queries"
                ) as progress:
                    self.encoder.eval()
                    for _ in range(available_cpus()):
                        worker = asyncio.create_task(
                            aio_search_worker(progress, results, queue)
                        )
                        workers.append(worker)

                    batcher = self.batcher.initialize(self.batchsize)
                    with torch.no_grad():
                        for batch in batchiter(self.batchsize, items):
                            await batcher.aio_reduce(batch, reducer, queue)

                    # Just wait for this to end
                    await queue.join()

            finally:
                # Stop all retriever workers
                for worker in workers:
                    worker.cancel()
            return results

        logger.info("Retrieve all with %d CPUs", available_cpus())
        results = asyncio.run(aio_process())
        return results

    def retrieve(self, query: TopicRecord, top_k=None) -> List[ScoredDocument]:
        """Search with document-at-a-time (DAAT) strategy

        :param top_k: Overrides the default top-K value
        """

        # Build up iterators
        vector = self.encoder([query]).value[0].cpu().detach().numpy()
        (ix,) = vector.nonzero()  # ix represents the position without 0 in the vector
        query = {
            ix: float(v) for ix, v in zip(ix, vector[ix])
        }  # generate a dict: {position:value}
        return self.index.retrieve(
            query, top_k or self.topk, **self.get_retrieval_options()
        )

    def get_retrieval_options(self) -> dict[str, Any]:
        """Returns extra retrieval options to be used when retrieving"""
        return {}

    def __validate__(self):
        # Checks that we are using the right retriever
        assert isinstance(
            self, self.index.Retriever
        ), f"{type(self)} is not an instance of {self.index.Retriever}"


class AbstractSparseRetrieverIndex(Config, ABC):
    Retriever = SparseRetriever

    documents: Param[DocumentStore]
    """The indexed document collection"""

    index: impact_index.Index
    ordered = False

    def initialize(self, in_memory: bool):
        self.index = impact_index.Index.load(str(self.index_path.absolute()), in_memory)

    @abstractmethod
    def retrieve(
        self, query: Dict[int, float], top_k: int, **kwargs
    ) -> List[ScoredDocument]:
        ...

    @abstractmethod
    async def aio_retrieve(
        self, query: Dict[int, float], top_k: int, **kwargs
    ) -> List[ScoredDocument]:
        ...


# ---
# --- CIFF file
# ---


class CIFFBuilder:
    def __init__(self, path: Path, levels: int):
        """Constructs a CIFF index

        :param path: _description_
        :param levels: _description_
        """
        # Path of the final index ("file.ciff")
        self.path = path


# ---
# --- Sparse index with the impact_index library
# ---


class SparseRetrieverIndex(AbstractSparseRetrieverIndex):
    index_path: Meta[Path]

    ordered = False

    index: impact_index.Index

    def initialize(self, in_memory: bool):
        self.index = impact_index.Index.load(str(self.index_path.absolute()), in_memory)

    def retrieve(self, query: Dict[int, float], top_k: int) -> List[ScoredDocument]:
        results = []
        for sd in self.index.search_maxscore(query, top_k):
            results.append(
                ScoredDocument(
                    self.documents.document_int(sd.docid),
                    sd.score,
                )
            )

        return results

    async def aio_retrieve(
        self, query: Dict[int, float], top_k: int
    ) -> List[ScoredDocument]:
        results = []
        for sd in await self.index.aio_search_maxscore(query, top_k):
            results.append(
                ScoredDocument(
                    self.documents.document_int(sd.docid),
                    sd.score,
                )
            )

        return results


class SparseRetrieverIndexBuilder(AbstractSparseRetrieverIndexBuilder[InputType]):
    in_memory: Meta[bool] = False
    """Whether the index should be fully loaded in memory (otherwise, uses
    virtual memory)"""

    index_path: Meta[Path] = field(default_factory=PathGenerator("index"))

    checkpoint_frequency: Meta[int] = 0
    """Checkpoint frequency (allows recovery at the cost of writing some
    information to disk)"""

    max_postings: Meta[Optional[int]] = None
    """Number of postings before dumping a term postings to disk"""

    def task_outputs(self, dep):
        """Returns a sparse retriever index that can be used by a
        SparseRetriever to search efficiently for documents"""

        return dep(
            SparseRetrieverIndex.C(index_path=self.index_path, documents=self.documents)
        )

    def create_index_builder(self):
        if self.index_path.is_dir() and self.checkpoint_frequency == 0:
            # Removing partially built index
            shutil.rmtree(self.index_path)
        self.index_path.mkdir(parents=True, exist_ok=True)

        # Setup options
        options = impact_index.BuilderOptions()
        options.checkpoint_frequency = self.checkpoint_frequency
        if self.max_postings is not None:
            options.in_memory_threshold = self.max_postings

        # and create the index builder
        self.indexer = impact_index.IndexBuilder(str(self.index_path), options)
        return self.indexer.get_checkpoint_doc_id()

    def build_index(self):
        self.indexer.build(self.in_memory)

    def add_encoded_document(self, docid, encoded, nonzero_ix):
        self.indexer.add(
            docid,
            nonzero_ix.astype(np.uint64),
            encoded.value[nonzero_ix],
        )


# ---
# --- Sparse index with the BMP library
# --- https://github.com/pisa-engine/BMP
# ---


class BMPSparseRetriever(SparseRetriever):
    """A Block-Max Pruning retriever"""

    alpha: Param[float]
    """Granularity of approximation (0 to 1, 1 = no approximation)"""

    beta: Param[float]
    """Percentage of query tokens to keep (0 to 1, 1 = no pruning)"""

    def get_retrieval_options(self) -> dict[str, Any]:
        """Returns extra retrieval options to be used when retrieving"""
        return {"alpha": self.alpha, "beta": self.beta}



class BMPSparseRetrieverIndex(AbstractSparseRetrieverIndex):
    Retriever = BMPSparseRetriever

    index_path: Meta[Path]
    """The path of the BMP index"""

    index: impact_index.Index

    def initialize(self, in_memory: bool):
        if not in_memory:
            logger.warning("BMP indices are in-memory only")
        try:
            from bmp import Searcher
        except ModuleNotFoundError:
            logger.warning(
                "Did not find the Block Max Pruning (bmp) library. "
                "Check https://github.com/pisa-engine/BMP"
            )
            raise
        logger.info("Loading BMP index from %s", self.index_path)
        self.searcher = Searcher(str(self.index_path))

    def retrieve(
        self, query: Dict[int, float], top_k: int, alpha=None, beta=None
    ) -> List[ScoredDocument]:
        results = []
        doc_ids, scores = self.searcher.search(
            {str(ix): float(value) for ix, value in query.items()},
            k=top_k,
            alpha=alpha,
            beta=beta,
        )

        for doc_id, score in zip(doc_ids, scores):
            results.append(
                ScoredDocument(
                    self.documents.document_int(int(doc_id)),
                    score,
                )
            )

        return results

    async def aio_retrieve(
        self, query: Dict[int, float], top_k: int, **kwargs
    ) -> List[ScoredDocument]:
        return self.retrieve(query, top_k, **kwargs)

class Sparse2BMPConverter(Task):
    index: Param[SparseRetrieverIndex]
    """The sparse index"""

    bmp_index_path: Meta[Path] = field(default_factory=PathGenerator("index.bmp"))
    """The final index path"""

    block_size: Param[int]
    """The block size"""

    compress_range: Param[bool]
    """Compress the BM index"""

    def task_outputs(self, dep):
        """Returns a sparse retriever index that can be used by a
        SparseRetriever to search efficiently for documents"""

        return dep(
            BMPSparseRetrieverIndex.C(
                index_path=self.bmp_index_path, documents=self.index.documents
            )
        )

    def execute(self):
        logging.info("Loading the index")
        self.index.initialize(False)

        logging.info("Converting to BMP")
        self.index.index.to_bmp(str(self.bmp_index_path), self.block_size, self.compress_range)

        logging.info("Done")


class BMPSparseRetrieverIndexBuilder(SparseRetrieverIndexBuilder[InputType]):
    """Index using a BMP index
    """
    block_size: Param[int]
    """The block size"""

    compress_range: Param[bool]
    """Compress the BM index"""

    bmp_index_path: Meta[Path] = field(default_factory=PathGenerator("index.bmp"))
    """The final index path"""

    def task_outputs(self, dep):
        """Returns a sparse retriever index that can be used by a
        SparseRetriever to search efficiently for documents"""

        return dep(
            BMPSparseRetrieverIndex.C(
                index_path=self.bmp_index_path, documents=self.documents
            )
        )

    def build_index(self):
        # Build the index
        logger.info("Building the index")
        index = self.indexer.build(False)

        logger.info("Converting to BMP index")
        index.to_bmp(str(self.bmp_index_path), self.block_size, self.compress_range)

        # Removes the old index
        logger.info("Removing the old index path")
        index = None
        shutil.rmtree(self.index_path, ignore_errors=True)



import logging
from pathlib import Path

import pytest
from experimaestro import ObjectStore
from experimaestro.xpmutils import DirectoryContext
from datamaestro_text.data.ir import TextItem, IDItem
from xpmir.documents.samplers import HeadDocumentSampler
from xpmir.index.faiss import FaissRetriever, IndexBackedFaiss
from xpmir.test.utils.utils import SampleDocumentStore, SparseRandomTextEncoder

indexspecs = ["Flat", "HNSW"]


@pytest.mark.parametrize("indexspec", indexspecs)
def test_faiss_indexation(tmp_path: Path, indexspec):
    # Experimaestro context
    from experimaestro.taskglobals import Env

    Env.taskpath = tmp_path / "task"
    Env.taskpath.mkdir()
    context = DirectoryContext(tmp_path)
    objects = ObjectStore()

    # Build the FAISS index
    documents = SampleDocumentStore(num_docs=100)
    sampler = HeadDocumentSampler(documents=documents, max_ratio=0.5)
    encoder = SparseRandomTextEncoder(dim=1000, sparsity=0.0)
    builder = IndexBackedFaiss(
        indexspec=indexspec,
        encoder=encoder,
        normalize=False,
        sampler=sampler,
        documents=documents,
    )
    builder_instance = builder.instance(context=context, objects=objects)
    builder_instance.execute()

    # Retrieve with the index
    topk = 10
    retriever = FaissRetriever(encoder=encoder, topk=topk, index=builder).instance(
        context=context, objects=objects
    )
    retriever.initialize()

    documents = builder_instance.documents.documents
    x_docs = retriever.encoder([d[TextItem].text for d in documents.values()]).value
    scores = x_docs @ x_docs.T

    for ix, document in enumerate(documents.values()):
        scoredDocuments = retriever.retrieve(document)
        scoredDocuments.sort(reverse=True)

        expected = list(scores[ix].sort(descending=True).indices[:topk].numpy())
        logging.debug("%s vs %s", scores[ix], scoredDocuments)
        observed = [int(sd.document[IDItem].id) for sd in scoredDocuments]

        assert expected == observed



from omegaconf import MISSING  # noqa: F401
from xpmir.letor import Random  # noqa: F401

from xpmir.learning.devices import Device, CudaDevice  # noqa: F401
from xpmir.papers import configuration, attrs_cached_property  # noqa: F401

from xpmir.experiments.helpers.launchers import LauncherSpecification  # noqa: F401
from xpmir.experiments.helpers import PaperExperiment, NeuralIRExperiment  # noqa: F401



from abc import ABC, abstractmethod
from typing import List, Union, Sequence
from attrs import evolve

import torch
from attrs import define
from experimaestro import Config


@define
class SimilarityInput(Sequence["SimilarityInput"]):
    value: torch.Tensor
    """A 3D tensor (batch x max_length x dim)"""

    mask: torch.BoolTensor
    """Mask for tokens of shape (batch x max_length)"""

    def __getitem__(self, index: Union[int, slice]) -> "SimilarityInput":
        return SimilarityInput(self.value[index], self.mask[index])

    def __len__(self) -> int:
        return len(self.value)


@define
class SimilarityInputWithTokens(SimilarityInput):
    tokens: List[List[str]]
    """A 3D tensor (batch x max_length x dim)"""

    def __getitem__(self, index: Union[int, slice]) -> "SimilarityInput":
        return SimilarityInputWithTokens(
            self.value[index], self.mask[index], self.tokens[index]
        )


@define
class SimilarityOutput(ABC):
    """Output for token similarities"""

    similarity: torch.Tensor
    """Similarity of each token

    The shape (Bq x Lq x Bd x Ld) when computing products, or (B x Lq x Ld) when
    computing pairs"""

    @abstractmethod
    def q_view(self, x: torch.Tensor) -> torch.Tensor:
        ...

    @abstractmethod
    def d_view(self, x: torch.Tensor) -> torch.Tensor:
        ...


@define
class PairsSimilarityOutput(SimilarityOutput):
    def q_view(self, x: torch.Tensor) -> torch.Tensor:
        return x.unsqueeze(2)

    def d_view(self, x: torch.Tensor) -> torch.Tensor:
        return x.unsqueeze(1)


@define
class ProductSimilarityOutput(SimilarityOutput):
    def d_view(self, x: torch.Tensor) -> torch.Tensor:
        return x.reshape(1, 1, *x.shape)

    def q_view(self, x: torch.Tensor) -> torch.Tensor:
        return x.reshape(*x.shape, 1, 1)


class Similarity(Config, ABC):
    """Base class for similarity between two texts representations (3D tensor
    batch x length x dim)"""

    def preprocess(self, encoded: SimilarityInput) -> SimilarityInput:
        """Optional preprocessing"""
        return encoded

    @abstractmethod
    def compute_pairs(
        self,
        queries: SimilarityInput,
        documents: SimilarityInput,
    ) -> SimilarityOutput:
        """Computes the score of queries and documents

        :param queries: Queries as a (B1 x L1 x dim) tensor
        :param documents: Documents as a (B2 x L2 x dim) tensor with B2 = B1 if
            product is False
        :param product: Computes the scores between all queries and documents
        :return: The score of all topics/documents (B1 x B2)
        """
        ...

    @abstractmethod
    def compute_product(
        self,
        queries: SimilarityInput,
        documents: SimilarityInput,
    ) -> SimilarityOutput:
        """Computes the score of queries and documents

        :param queries: Queries as a (B x L1 x dim) tensor
        :param documents: Documents as a (B x L2 x dim) tensor with B2 = B1 if
            product is False
        :param product: Computes the scores between all queries and documents
        :return: The similarity of query/document pairs (B)
        """
        ...


class DotProductSimilarity(Similarity):
    def compute_product(
        self,
        queries: SimilarityInput,
        documents: SimilarityInput,
    ):
        q = queries.value
        d = documents.value

        # Bq x Lq x Bd x Ld
        inner = (q.flatten(0, 1) @ d.flatten(0, 1).transpose(0, 1)).reshape(
            q.shape[:2] + d.shape[:2]
        )

        # Max on document tokens, sum over query tokens
        return ProductSimilarityOutput(inner)

    def compute_pairs(
        self,
        queries: SimilarityInput,
        documents: SimilarityInput,
    ):
        q = queries.value
        d = documents.value

        # B x Lq x Ld
        assert len(q) == len(d), "Batch sizes don't match"
        inner = q @ d.transpose(1, 2)

        return PairsSimilarityOutput(inner)


class CosineSimilarity(DotProductSimilarity):
    """Cosine similarity between two texts representations (3D tensor batch x
    length x dim)"""

    def preprocess(self, output: SimilarityInput):
        value = output.value / output.value.norm(p="fro", dim=-1, keepdim=True)
        return evolve(output, value=value)



import torch
from typing import Tuple
from experimaestro import Param
from datamaestro_text.data.ir import TextItem
from xpmir.distributed import DistributableModel
from xpmir.learning.batchers import Batcher
from xpmir.learning.context import TrainerContext
from xpmir.letor.records import (
    BaseRecords,
    PairwiseRecords,
)
from xpmir.rankers import LearnableScorer
from xpmir.text import TokenizerOptions
from xpmir.text.encoders import TextEncoderBase, TripletTextEncoder
from xpmir.rankers import (
    DuoLearnableScorer,
    DuoTwoStageRetriever,
    Retriever,
)
from xpmir.utils.utils import easylog

logger = easylog()


class CrossScorer(LearnableScorer, DistributableModel):
    """Query-Document Representation Classifier

    Based on a query-document representation representation (e.g. BERT [CLS] token).
    AKA Cross-Encoder
    """

    max_length: Param[int]
    """Maximum length (in tokens) for the query-document string"""

    encoder: Param[TextEncoderBase[Tuple[str, str], torch.Tensor]]
    """an encoder for encoding the concatenated query-document tokens which
    doesn't contains the final linear layer"""

    def __validate__(self):
        super().__validate__()
        assert not self.encoder.static(), "The vocabulary should be learnable"

    def __initialize__(self, options):
        super().__initialize__(options)
        self.encoder.initialize(options)
        self.classifier = torch.nn.Linear(self.encoder.dimension, 1)
        self.tokenizer_options = TokenizerOptions(max_length=self.max_length)

    def forward(self, inputs: BaseRecords, info: TrainerContext = None):
        # Encode queries and documents
        pairs = self.encoder(
            [
                (tr[TextItem].text, dr[TextItem].text)
                for tr, dr in zip(inputs.topics, inputs.documents)
            ],
            # options=self.tokenizer_options,
        )  # shape (batch_size * dimension)
        return self.classifier(pairs.value).squeeze(1)

    def distribute_models(self, update):
        self.encoder = update(self.encoder)


class DuoCrossScorer(DuoLearnableScorer, DistributableModel):
    """Preference based classifier

    This scorer can be used to train a DuoBERT-type model.
    """

    encoder: Param[TripletTextEncoder]
    """The encoder to use for the Duobert model"""

    def __validate__(self):
        assert not self.encoder.static(), "The vocabulary should be learnable"

    def __initialize__(self, options):
        super().__initialize__(options)
        self.encoder.initialize(options)
        self.classifier = torch.nn.Linear(self.encoder.dimension, 1)

    def forward(self, inputs: PairwiseRecords, info: TrainerContext = None):
        """Encode the query-document-document"""
        triplets = self.encoder(
            [
                (q.text, d_1.text, d_2.text)
                for q, d_1, d_2 in zip(
                    inputs.unique_queries, inputs.positives, inputs.negatives
                )
            ]
        )
        return self.classifier(triplets).squeeze(1)

    def distribute_models(self, update):
        self.encoder.model = update(self.encoder.model)

    def getRetriever(
        self,
        retriever: "Retriever",
        batch_size: int,
        batcher: Batcher = Batcher(),
        top_k=None,
        device=None,
    ):
        """The given the base_retriever and return a two stage retriever
        specialized for Duobert"""
        return DuoTwoStageRetriever(
            retriever=retriever,
            scorer=self,
            batchsize=batch_size,
            batcher=batcher,
            device=device,
            top_k=top_k,
        )



import re
from experimaestro.mkdocs.metaloader import DependencyInjectorFinder

DependencyInjectorFinder.install(
    re.compile(r"^(torch|faiss|pandas|bs4|pytorch_transformers|pytrec_eval|apex)($|\.)")
)



from pathlib import Path
from attrs import define
from typing import Dict, Optional, Union
from xpmir.rankers import Scorer
from xpmir.evaluation import EvaluationsCollection
from xpmir.text.huggingface import MLMEncoder


@define(kw_only=True)
class PaperResults:
    models: Dict[str, Union[Scorer, MLMEncoder]]
    """List of models with their identifier"""

    evaluations: EvaluationsCollection
    """The evaluation results"""

    tb_logs: Optional[Dict[str, Path]]
    """Tensorboard directory for each model"""



import click
from pathlib import Path
import os
from datamaestro import prepare_dataset
import logging

from experimaestro import experiment
from xpmir.evaluation import Evaluate
from xpmir.rankers.standard import BM25
from xpmir.interfaces.anserini import AnseriniRetriever, IndexCollection


logging.basicConfig(level=logging.INFO)
CPU_COUNT = len(os.sched_getaffinity(0))

# --- Defines the experiment


@click.option("--debug", is_flag=True, help="Print debug information")
@click.option("--port", type=int, default=12345, help="Port for monitoring")
@click.option("--dataset", default="gov.nist.trec.adhoc.1")
@click.argument("workdir", type=Path)
@click.command()
def cli(port, workdir, dataset, debug):
    """Runs an experiment"""
    logging.getLogger().setLevel(logging.DEBUG if debug else logging.INFO)

    bm25 = BM25()

    # Sets the working directory and the name of the xp
    with experiment(workdir, "bm25", port=port) as xp:
        # Index the collection
        xp.setenv("JAVA_HOME", os.environ["JAVA_HOME"])
        ds = prepare_dataset(dataset)

        documents = ds.documents
        index = IndexCollection(
            documents=documents,
            storePositions=True,
            storeDocvectors=True,
            storeContents=True,
            threads=CPU_COUNT,
        ).submit()

        # Search with BM25
        bm25_retriever = AnseriniRetriever(k=1500, index=index, model=bm25).tag(
            "model", "bm25"
        )

        bm25_eval = Evaluate(dataset=ds, retriever=bm25_retriever).submit()

    logging.info("BM25 results on TREC 1")
    logging.info(bm25_eval.results.read_text())


if __name__ == "__main__":
    cli()



from datamaestro import prepare_dataset

# Returns Documents object
documents = prepare_dataset("irds.antique.train").documents

for _, document in zip(range(20), documents.iter()):
    print(document.docid, document.text)  # (qid, text) tuple, noqa: T201



from abc import ABC, abstractmethod
from functools import cached_property
from typing import List
from pathlib import Path
from experimaestro import Config, Meta


class IDList(Config, ABC):
    """A configuration that returns a list of ids"""

    @property
    @abstractmethod
    def ids(self) -> List[str]:
        """Returns the list of IDs"""
        ...


class FileIDList(IDList):
    """A file-based list of IDs"""

    path: Meta[Path]

    @cached_property
    def ids(self) -> List[str]:
        return list(self.path.read_text().split("\n"))



from typing import Callable, DefaultDict, Dict, List, Type, TypeVar
from experimaestro import Config


class Hook(Config):
    """Base class for all hooks"""

    pass


HookType = TypeVar("HookType")


class InitializationHook(Hook):
    """Base class for hooks before/after initialization"""

    def after(self, context: "Context"):
        """Called after initialization"""
        pass

    def before(self, context: "Context"):
        """Called before initialization"""
        pass


class Context:
    """Generic computational context"""

    hooksmap: Dict[Type, List[Hook]]
    """Map of hooks"""

    def __init__(self, hooks: List[Hook] = []):
        self.hooksmap = DefaultDict(lambda: [])
        for hook in hooks:
            self.add_hook(hook)

    def hooks(self, cls: Type[HookType]) -> List[HookType]:
        """Returns all the hooks"""
        return self.hooksmap.get(cls, [])  # type: ignore

    def call_hooks(self, cls: Type, method: Callable, *args, **kwargs):
        for hook in self.hooks(cls):
            method(hook, *args, **kwargs)

    def add_hook(self, hook):
        for cls in hook.__class__.__mro__:
            self.hooksmap[cls].append(hook)



"""Utility functions when aligning two tokenizations"""

from typing import (
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Protocol,
    Iterable,
    Tuple,
    TypeVar,
)
from dataclasses import dataclass


class TokenWithRange(Protocol):
    """A token with its range"""

    string: str
    start: int
    end: int


QueryToken = TypeVar("QueryToken", bound=TokenWithRange)
DocumentToken = TypeVar("DocumentToken", bound=TokenWithRange)


@dataclass
class Match(Generic[QueryToken, DocumentToken]):
    string: str
    query: List[QueryToken]
    document: List[DocumentToken]


def findmatches(
    query_toks: Iterator[QueryToken], document_toks: Iterator[DocumentToken], all=True
) -> Dict[str, Match[QueryToken, DocumentToken]]:
    """Find query tokens in documents"""
    # Find all words in query
    matching = {}

    for token in query_toks:
        if all or token.string not in matching:
            match = matching.setdefault(token.string, Match(token.string, [], []))
            match.query.append(token)

    # Find all query words in documents
    for token in document_toks:
        match = matching.get(token.string, None)
        if match is not None:
            if all or len(match.document) == 0:
                match.document.append(token)

    return matching


def optnext(it: Iterator[TokenWithRange]) -> Optional[TokenWithRange]:
    try:
        return next(it)
    except StopIteration:
        return None


def findpositions(
    keys: Iterator[QueryToken], tokens: Iterator[DocumentToken]
) -> Iterator[Tuple[QueryToken, DocumentToken]]:
    """Find the position of `keys` in `tokens`

    supposes that:

    - keys and tokens are sorted, i.e. start_{i+1} >= end_i
    - all the keys should appear in tokens


    Returns tuples (c, )
    """

    c = optnext(keys)
    if c is None:
        return

    for token in tokens:
        # no match to be done (query)
        if c is None:
            break

        if token.start <= c.start and token.end > c.start:
            yield (c, token)
            c = optnext(keys)

    assert c is None, "Some tokens have not been found"



from torch.optim.lr_scheduler import LambdaLR
from experimaestro import Config, Param


class Scheduler(Config):
    """Base class for all optimizers schedulers"""

    def __call__(self, optimizer, num_training_steps: int, *, last_epoch=-1, **kwargs):
        raise NotImplementedError(f"Not implemented in {self.__class__}")


class LinearWithWarmup(Scheduler):
    """Linear warmup followed by decay"""

    num_warmup_steps: Param[int]
    """Number of warmup steps"""

    min_factor: Param[float] = 0.0
    """Minimum multiplicative factor"""

    def lr_lambda(self, current_step: int, num_training_steps: int):
        # Still warming up
        if current_step < self.num_warmup_steps:
            return float(current_step) / float(max(1, self.num_warmup_steps))

        # Not warming up: the ratio is between 1 (after warmup) and 0 (at the end)
        factor = max(
            0.0,
            float(num_training_steps - current_step)
            / float(max(1, num_training_steps - self.num_warmup_steps)),
        )

        # Shift/scale so it is between 1 and min factor
        return (factor + self.min_factor) / (1.0 + self.min_factor)

    def __call__(self, optimizer, num_training_steps, *, last_epoch=-1):
        return LambdaLR(
            optimizer,
            lambda current_step: self.lr_lambda(current_step, num_training_steps),
            last_epoch=last_epoch,
        )


class CosineWithWarmup(Scheduler):
    """Cosine schedule with warmup

    Uses the implementation of the transformer library

    https://huggingface.co/docs/transformers/main_classes/optimizer_schedules#transformers.get_cosine_schedule_with_warmup

    """

    num_warmup_steps: Param[int]
    """Number of warmup steps"""

    num_cycles: Param[float] = 0.5
    """Number of cycles"""

    def __call__(self, optimizer, num_training_steps, *, last_epoch=-1):
        import transformers

        return transformers.get_cosine_schedule_with_warmup(
            optimizer,
            self.num_warmup_steps,
            num_training_steps,
            self.num_cycles,
            last_epoch=last_epoch,
        )



from typing import Optional, List
from experimaestro import Param
from xpmir.utils.utils import foreach
from . import Retriever


class SumRetriever(Retriever):
    """Combines the scores of various retrievers"""

    retrievers: Param[List[Retriever]]
    """retrievers"""

    weights: Param[Optional[List[int]]] = None
    """The weights of the retrievers"""

    def __validate__(self):
        super().__validate__()
        assert self.weights is None or len(self.weights) == len(self.retrievers)

    def initialize(self):
        super().initialize()
        foreach(self.retrievers, lambda retriever: retriever.initialize())



import re
from typing import Optional
from abc import ABC, abstractmethod
import torch.nn as nn
from typing import Iterator, NamedTuple
from experimaestro import Param, Config, PathSerializationLWTask
import torch
from xpmir.learning.optim import Module, ModuleLoader, ModuleInitMode
import logging

logger = logging.getLogger("xpmir.learning")


class ParameterElement(NamedTuple):
    name: str
    """Name of the parameter (with respect to the main model)

    See
    https://pytorch.org/docs/stable/generated/torch.nn.Module.html#torch.nn.Module.named_parameters
    """

    module: nn.Module
    """The torch module"""

    parameter: nn.parameter.Parameter
    """The parameter object"""

    selected: bool
    """Selection status"""

    def set(self, value: nn.parameter.Parameter):
        """Set this parameter to a new value"""
        *parts, last = self.name.split(".")
        module = self.module
        for part in parts:
            module = getattr(module, part)
        setattr(module, last, value)

    def __repr__(self):
        return (
            f"(name={self.name}, id(parameter)={id(self.parameter)},"
            f" selected={self.selected})"
        )


class ParametersIterator(Config, ABC):
    """Iterator over module parameters

    This can be useful to freeze some layers, or perform any other
    parameter-wise operation
    """

    @abstractmethod
    def iter(self) -> Iterator[ParameterElement]:
        """Iterate over parameters

        :yield: An iterator over triplets (name of the parameter, parameters,
            selected or not)
        """
        ...

    def selected(self) -> Iterator[ParameterElement]:
        """Iterates over selected parameters"""
        for element in self.iter():
            if element.selected:
                yield element


class RegexParametersIterator(ParametersIterator):
    """Itertor over all the parameters which match the given regex"""

    negative_regex: Param[str] = ""
    """The negative regex expression (should not match)"""

    regex: Param[str]
    """The regex expression"""

    model: Param[Module]
    """The model we want to select the parameters from"""

    def __post_init__(self):
        self._regex = re.compile(self.regex) if self.regex else None
        self._negative_regex = (
            re.compile(self.negative_regex) if self.negative_regex else None
        )

    def should_pick(self, name: str) -> bool:
        """given the name of the str, return true if the regex expression
        matches"""
        if self.negative_regex:
            if self._negative_regex.search(name):
                return False
        if self._regex:
            value = bool(self._regex.search(name))
            return value
        return True

    def iter(self) -> Iterator[ParameterElement]:
        for name, parameters in self.model.named_parameters():
            yield ParameterElement(name, self.model, parameters, self.should_pick(name))


class InverseParametersIterator(ParametersIterator):
    """Inverse the selection of a parameter iterator"""

    iterator: Param[ParametersIterator]

    def iter(self) -> Iterator[ParameterElement]:
        yield from (
            ParameterElement(name, module, param, not (selected))
            for name, module, param, selected in self.iterator.iter()
        )


class SubParametersIterator(ParametersIterator):
    """Wraps a parameter iterator over a global model and a selector
    over a subpart of the model"""

    model: Param[Module]
    """The model from which the parameters should be gathered"""

    iterator: Param[ParametersIterator]
    """The sub-model iterator"""

    default: Param[bool]
    """Default value for parameters not within the sub-model"""

    def iter(self) -> Iterator[ParameterElement]:
        # Gather all the model parameters
        model_params = {
            id(p): ParameterElement(name, self.model, p, self.default)
            for name, p in self.model.named_parameters()
        }

        # Copy the selection status from the sub-model iterator
        for element in self.iterator.iter():
            id_param = id(element.parameter)
            
            if mp := model_params.get(id_param, None):
                if mp.selected != element.selected:
                    model_params[id_param] = ParameterElement(
                        mp.name, mp.module, mp.parameter, element.selected
                    )
            else:
                raise RuntimeError("Sub-model parameters are not model parameters")

        # Return everything
        yield from model_params.values()


class NameMapper(Config, ABC):
    """Changes name of parameters"""

    @abstractmethod
    def __call__(self, source: str) -> str:
        ...


class PrefixRenamer(NameMapper):
    """Changes name of parameters"""

    model: Param[str]
    """Prefix in model"""

    data: Param[str]
    """Prefix in data"""

    def __call__(self, source: str) -> str:
        if source.startswith(self.model):
            return f"{self.data}{source[len(self.model):]}"

        return source


class PartialModuleLoader(PathSerializationLWTask):
    """Allows to load only a part of the parameters"""

    selector: Param[ParametersIterator]
    """The selectors gives the list of parameters for which some"""

    mapper: Param[Optional[NameMapper]] = None
    """Maps parameter names so it matches so the saved ones"""

    def execute(self):
        """Combine the model in the selectors"""
        self.value.initialize(ModuleInitMode.NONE.to_options())
        data = torch.load(self.path)
        logger.info(
            "(partial module loader) Loading parameters from %s into %s",
            self.path,
            type(self.value).__name__,
        )

        partial_data = {}
        value_names = set(key for key, _ in self.value.named_parameters())
        for name, _, selected in self.selector.iter():
            if selected:
                if self.mapper is None:
                    key = name
                    logger.debug(f"Selected: {name}")
                else:
                    key = self.mapper(name)
                    logger.debug(f"Selected: {key} -> {name}")

                assert key in data, (
                    f"{key} is not in loaded parameters:" f"{', '.join(data.keys())}"
                )
                partial_data[name] = data[key]

        # Log some potentially useful information
        data_names = set(partial_data.keys())
        inter_names = value_names.intersection(data_names)
        not_used = data_names.difference(inter_names)

        if len(not_used) > 0:
            logger.error("Some selected parameters are not model parameters")
            logger.error("Unused parameters: %s", ", ".join(not_used))

            logger.error("Model parameters: %s", ", ".join(value_names))
            logger.error("Data parameters: %s", ", ".join(data.keys()))
            raise RuntimeError("Some selected parameters are not model parameters")

        assert len(inter_names) > 0, "No common parameters?"

        # Loads with strict False since some keys might not be
        # in the data
        self.value.load_state_dict(partial_data, strict=False)

    @staticmethod
    def from_module_loader(
        module_loader: "ModuleLoader",
        value: Config,
        selector: ParametersIterator,
        *,
        mapper: Optional[NameMapper] = None,
    ):
        """Creates a new partial module loader from another module loader

        :param module_loader: The module loader
        :param value: The module for which parameters should be loaded
        :param selector: The selector to restrict the set of parameters
        :return: A new partial module loader
        """
        assert (
            module_loader.__xpm__.task is not None
        ), "No task associated with the module_loader"

        pml = PartialModuleLoader(
            selector=selector, value=value, path=module_loader.path, mapper=mapper
        )
        pml.copy_dependencies(module_loader)
        return pml


class SubModuleLoader(PathSerializationLWTask):
    """Allows to load only a part of the parameters (with automatic renaming)"""

    selector: Param[ParametersIterator]
    """The selectors gives the list of parameters for which
    loaded parameters should be used"""

    saved_value: Param[Optional[Module]] = None
    """The original module that is being loaded (optional,
    allows to map names)"""

    def execute(self):
        """Combine the model in the selectors"""
        self.value.initialize(None)
        data = torch.load(self.path)
        logger.info(
            "(partial module loader) Loading parameters from %s into %s",
            self.path,
            type(self.value).__name__,
        )

        # Creates a mapper if needed
        mapper = None
        if self.saved_value:
            mapper = {
                id(params): key for key, params in self.saved_value.named_parameters()
            }

        partial_data = {}
        value_names = set(key for key, _ in self.value.named_parameters())
        for element in self.selector.iter():
            name, params, selected = element.name, element.parameter, element.selected
            if selected:
                if mapper is None:
                    key = name
                    logger.debug(f"Selected: {name}")
                else:
                    key = mapper[id(params)]
                    logger.debug(f"Selected: {key} -> {name}")

                assert key in data, (
                    f"{key} is not in loaded parameters:" f"{', '.join(data.keys())}"
                )
                partial_data[name] = data[key]

        # Log some potentially useful information
        data_names = set(partial_data.keys())
        inter_names = value_names.intersection(data_names)
        not_used = data_names.difference(inter_names)

        if len(not_used) > 0:
            logger.error("Some selected parameters are not model parameters")
            logger.error("Unused parameters: %s", ", ".join(not_used))

            logger.error("Model parameters: %s", ", ".join(value_names))
            logger.error("Data parameters: %s", ", ".join(data.keys()))
            raise RuntimeError("Some selected parameters are not model parameters")

        assert len(inter_names) > 0, "No common parameters?"

        # Loads with strict False since some keys might not be
        # in the data
        self.value.load_state_dict(partial_data, strict=False)

    @staticmethod
    def from_module_loader(
        module_loader: "ModuleLoader",
        saved_value: Optional[Config],
        value: Config,
        selector: ParametersIterator,
    ):
        """Creates a new partial module loader from another module loader

        :param module_loader: The module loader
        :param saved_value: The configuration which has been saved
        :param value: The module for which parameters should be loaded (should
            be a sub-module of saved_value). If None, uses the module_loader
            value
        :param selector: The selector to restrict the set of parameters
        :return: A new partial module loader
        """
        assert (
            module_loader.__xpm__.task is not None
        ), "No task associated with the module_loader"

        assert isinstance(module_loader, ModuleLoader)

        pml = SubModuleLoader(
            selector=selector,
            value=value,
            path=module_loader.path,
            saved_value=module_loader.value if saved_value is None else saved_value,
        )
        pml.copy_dependencies(module_loader)
        return pml



from typing import List, Optional, Tuple, Dict, Any
from experimaestro import Param, Meta, tqdm
import torch
from datamaestro_text.data.ir import Documents
from xpmir.neural.dual import DualRepresentationScorer
from xpmir.learning.batchers import Batcher
from xpmir.learning import ModuleInitMode
from xpmir.letor import Device
from xpmir.letor.records import DocumentRecord, TopicRecord
from . import Retriever, ScoredDocument


class FullRetriever(Retriever):
    """Retrieves all the documents of the collection

    This can be used to build a small validation set on a subset of the
    collection - in that case, the scorer can be used through a
    TwoStageRetriever, with this retriever as the base retriever.
    """

    documents: Param[Documents]

    def retrieve(self, record: TopicRecord) -> List[ScoredDocument]:
        return [ScoredDocument(doc, 0.0) for doc in self.documents]


class FullRetrieverRescorer(Retriever):
    """Scores all the documents from a collection"""

    documents: Param[Documents]
    """The set of documents to consider"""

    scorer: Param[DualRepresentationScorer]
    """The scorer (a dual representation scorer)"""

    batchsize: Param[int] = 0
    batcher: Meta[Batcher] = Batcher()
    device: Meta[Optional[Device]] = None

    def initialize(self):
        self.query_batcher = self.batcher.initialize(self.batchsize)
        self.document_batcher = self.batcher.initialize(self.batchsize)
        self.scorer.initialize(ModuleInitMode.DEFAULT.to_options())

        # Compute with the scorer
        if self.device is not None:
            self.scorer.to(self.device.value)

    def _retrieve(
        self,
        batch: List[ScoredDocument],
        query: str,
        scoredDocuments: List[ScoredDocument],
    ):
        scoredDocuments.extend(self.scorer.rsv(query, batch))

    def encode_queries(self, queries: List[Tuple[str, str]], encoded: List[Any], pbar):
        """Encode queries and append the tensor of encoded queries to the encoded

        Args:
            queries (List[Tuple[str, str]]): The input queries (id/text)
            encoded (List[Tuple[List[str], torch.Tensor]]): Full list of topics ??
            it should be the List[torch.Tensor]
        """

        encoded.append(self.scorer.encode_queries([text for _, text in queries]))
        pbar.update(len(queries))
        return encoded

    def score(
        self,
        documents: List[DocumentRecord],
        queries: List,
        scored_documents: List[List[ScoredDocument]],
        pbar,
    ):
        """Score documents for a set of queries

        Every time the score process a batch of document together with whole set
        of queries

        scored_documents is filled with document batches, i.e. it contains [
        [s(q_0, d_0), ..., s(q_n, d0)], ..., [s(q_0, d_m), ..., s(q_n, d_m)] ]
        --> list of m*n

        :param documents: the batch of documents

        :param queries: List of queries

        :param scored_documents: (output) current lists of scored documents (one
            per query)
        """
        # Encode documents
        encoded = self.scorer.encode_documents(documents)

        # Process query by query
        new_scores = [[] for _ in documents]
        for ix in range(len(queries)):
            # Get a range of query records
            query = queries[ix : (ix + 1)]

            # Returns a query x document matrix
            scores = self.scorer.score_product(query.to(encoded.device), encoded, None)

            # Adds up to the lists
            scores = scores.flatten().detach()
            for ix, (document, score) in enumerate(zip(documents, scores)):
                new_scores[ix].append(ScoredDocument(document, float(score)))
                pbar.update(1)

        # Add each result to the full document list
        scored_documents.extend(new_scores)

    def retrieve(self, record: TopicRecord) -> List[ScoredDocument]:
        # Only use retrieve_all
        return self.retrieve_all({"_": record})["_"]

    def retrieve_all(
        self, queries: Dict[str, TopicRecord]
    ) -> Dict[str, List[ScoredDocument]]:
        """Input is a dictionary of query {id:text},
        return the a dictionary of {query_id: List of ScoredDocuments under the query}
        """

        self.scorer.eval()
        all_queries = list(queries.items())

        with torch.no_grad():
            # Encode all queries
            # each time the batcher will just encode a batchsize of queries
            # and then concat them together
            with tqdm(total=len(all_queries), desc="Encoding queries") as pbar:
                enc_queries = self.query_batcher.reduce(
                    all_queries, self.encode_queries, [], pbar
                )
            enc_queries = self.scorer.merge_queries(
                enc_queries
            )  # shape (len(queries), dimension)

            # Encode documents and score them
            scored_documents: List[List[ScoredDocument]] = []
            with tqdm(
                total=len(all_queries) * self.documents.documentcount,
                desc="Scoring documents",
            ) as pbar:
                self.document_batcher.process(
                    self.documents, self.score, enc_queries, scored_documents, pbar
                )

        qids = [qid for qid, _ in all_queries]
        return {qid: [sd[ix] for sd in scored_documents] for ix, qid in enumerate(qids)}



from abc import ABC, abstractmethod
from typing import Optional, Tuple, Iterator, Any
from experimaestro import Param, Config
import torch
import numpy as np
from datamaestro_text.data.ir import DocumentStore, TextItem, create_record
from xpmir.letor import Random
from xpmir.letor.records import DocumentRecord, PairwiseRecord, ProductRecords
from xpmir.letor.samplers import BatchwiseSampler, PairwiseSampler
from xpmir.utils.iter import RandomSerializableIterator, SerializableIterator


class DocumentSampler(Config, ABC):
    """How to sample from a document store"""

    documents: Param[DocumentStore]

    @abstractmethod
    def __call__(self) -> Tuple[Optional[int], Iterator[DocumentRecord]]:
        """Returns an indicative number of samples and an iterator"""
        raise NotImplementedError()

    def __iter__(self) -> Iterator[DocumentRecord]:
        """Shorthand method that directly returns an iterator"""
        _, iter = self()
        return iter


class HeadDocumentSampler(DocumentSampler):
    """A basic sampler that iterates over the first documents

    if max_count is 0, it iterates over all documents
    """

    max_count: Param[int] = 0
    """Maximum number of documents (if 0, no limit)"""

    max_ratio: Param[float] = 0
    """Maximum ratio of documents (if 0, no limit)"""

    def __call__(self) -> Tuple[int, Iterator[DocumentRecord]]:
        count = (self.max_ratio or 1) * self.documents.documentcount

        if self.max_count > 0:
            count = min(self.max_count, count)

        count = int(count)
        return count, self.iter(count)

    def iter(self, count):
        for _, document in zip(range(count), self.documents.iter_documents()):
            yield document


class RandomDocumentSampler(DocumentSampler):
    """A basic sampler that iterates over the first documents

    Either max_count or max_ratio should be non null
    """

    max_count: Param[int] = 0
    """Maximum number of documents (if 0, no limit)"""

    max_ratio: Param[float] = 0
    """Maximum ratio of documents (if 0, no limit)"""

    random: Param[Optional[Random]]
    """Random sampler"""

    def __call__(self) -> Tuple[int, Iterator[str]]:
        # Compute the number of documents to sample
        count = (self.max_ratio or 1) * self.documents.documentcount

        if self.max_count > 0:
            count = min(self.max_count, count)
        count = int(count)
        return count, self.iter(count)

    def iter(self, count) -> Iterator[str]:
        """Iterate over the documents"""
        state = np.random.RandomState() if self.random is None else self.random.state
        docids = state.choice(
            np.arange(self.documents.documentcount), size=count, replace=False
        )
        for docid in docids:
            yield self.documents.document_int(int(docid))


class RandomSpanSampler(BatchwiseSampler, PairwiseSampler):
    """This sampler uses positive samples coming from the same documents
    and negative ones coming from others

    Allows to (pre)-train as in co-condenser:
        L. Gao and J. Callan, “Unsupervised Corpus Aware Language Model
        Pre-training for Dense Passage Retrieval,” arXiv:2108.05540 [cs],
        Aug. 2021, Accessed: Sep. 17, 2021. [Online].
        http://arxiv.org/abs/2108.05540
    """

    documents: Param[DocumentStore]
    """The document store to use"""

    max_spansize: Param[int] = 1000
    """Maximum span size in number of characters"""

    def get_text_span(self, text, random):
        # return the two spans of text
        spanlen = min(self.max_spansize, len(text) // 2)

        max_start1 = len(text) - spanlen * 2
        start1 = random.randint(0, max_start1) if max_start1 > 0 else 0
        end1 = start1 + spanlen
        if start1 > 0 and text[start1 - 1] != " ":
            start1 = text.find(" ", start1) + 1
        if text[end1] != " ":
            end1 = text.rfind(" ", 0, end1)

        max_start2 = len(text) - spanlen
        start2 = random.randint(end1, max_start2) if max_start2 > end1 else end1
        end2 = start2 + spanlen
        if text[start2 - 1] != " ":
            start2 = text.find(" ", start2) + 1
        if text[end2 - 1] != " ":
            end2 = text.rfind(" ", 0, end2)

        # Rejet wrong samples
        if end2 <= start2 or end1 <= start1:
            return None

        return (text[start1:end1], text[start2:end2])

    def pairwise_iter(self) -> SerializableIterator[PairwiseRecord, Any]:
        def iter(random: np.random.RandomState):
            iter = self.documents.iter_sample(lambda m: random.randint(0, m))

            while True:
                record_pos_qry = next(iter)
                text_pos_qry = record_pos_qry[TextItem].text
                spans_pos_qry = self.get_text_span(text_pos_qry, random)

                record_neg = next(iter)
                text_neg = record_neg[TextItem].text
                spans_neg = self.get_text_span(text_neg, random)

                if not (spans_pos_qry and spans_neg):
                    continue

                yield PairwiseRecord(
                    create_record(text=spans_pos_qry[0]),
                    create_record(text=spans_pos_qry[1]),
                    create_record(text=spans_neg[random.randint(0, 2)]),
                )

        return RandomSerializableIterator(self.random, iter)

    def batchwise_iter(
        self, batch_size: int
    ) -> SerializableIterator[ProductRecords, Any]:
        def iterator(random: np.random.RandomState):
            # Pre-compute relevance matrix
            relevances = torch.diag(torch.ones(batch_size, dtype=torch.float))

            iter = self.documents.iter_sample(lambda m: random.randint(0, m))

            while True:
                batch = ProductRecords()
                while len(batch) < batch_size:
                    record = next(iter)
                    text = record.text
                    res = self.get_text_span(text, random)
                    if not res:
                        continue
                    batch.add_topics(create_record(text=res[0]))
                    batch.add_documents(create_record(text=res[1]))
                batch.set_relevances(relevances)
                yield batch

        return RandomSerializableIterator(self.random, iterator)



from abc import abstractmethod
from typing import Dict, Iterator, List, Optional
from experimaestro import Config, Param
import torch.nn as nn
import numpy as np
from xpmir.utils.utils import EasyLogger
from xpmir.learning import Module
from xpmir.learning.context import (
    TrainingHook,
    TrainerContext,
)

from xpmir.utils.utils import foreach


class Trainer(Config, EasyLogger):
    """Generic trainer"""

    hooks: Param[List[TrainingHook]] = []
    """Hooks for this trainer: this includes the losses, but can be adapted for
        other uses

        The specific list of hooks depends on the specific trainer
    """

    model: Param[Optional[Module]] = None
    """If the model to optimize is different from the model passsed to Learn,
    this parameter can be used – initialization is still expected to be done at
    the learner level"""

    def initialize(
        self,
        random: np.random.RandomState,
        context: TrainerContext,
    ):
        self.random = random

        # Generic style
        if self.model is None:
            self.model = context.state.model

        # Old style (to be deprecated)
        self.ranker = self.model

        self.context = context

        foreach(self.hooks, self.context.add_hook)

    def to(self, device):
        """Change the computing device (if this is needed)"""
        foreach(self.context.hooks(nn.Module), lambda hook: hook.to(device))

    @abstractmethod
    def iter_batches(self) -> Iterator:
        """Returns a (serializable) iterator over batches"""
        ...

    @abstractmethod
    def process_batch(self, batch):
        ...

    @abstractmethod
    def load_state_dict(self, state: Dict):
        ...

    @abstractmethod
    def state_dict(self):
        ...
